def load_and_preprocess_img(path):
    import numpy as np
    from PIL import Image
    from torchvision import transforms


    transform = transforms.Compose([
                    transforms.Resize((224, 224)),
                    transforms.ToTensor(),
                    transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                        std=[0.229, 0.224, 0.225])
                ])
    img = Image.open(path)
    img = transform(img).numpy()
    return np.expand_dims(img, axis=0)



from __future__ import annotations

import os, glob
import zipfile
from enum import Enum
import inspect
from pathlib import Path
from typing import List, Union
from onnc.bench.core.model.model import Model
from onnc.bench.core.dataset.dataset import Dataset
# from typing import TYPE_CHECKING

# if TYPE_CHECKING:
#     from . import Model

from . import ModelFormat


def _check_object_type(expect_type_str, obj):
    type_str = ""
    try:
        type_str += str(inspect.getmro((obj)))
    except:
        type_str += ""
    type_str += "|"
    type_str += str(type(obj))
    return expect_type_str in type_str


class IdentifierRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Identifier(metaclass=IdentifierRegistry):

    FORMAT = ModelFormat.NON_SPECIFIED

    @classmethod
    def is_me(cls, model: Model) -> bool:
        pass


def identify(model: Model) -> ModelFormat:
    for identifier in Identifier.REGISTRY:
        if identifier.is_me(model):
            return identifier.FORMAT
    raise NotImplementedError(f"Unable to identify {model.src}")


class H5(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = ModelFormat.H5

    @classmethod
    def is_me(cls, model: Model) -> bool:
        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False
        if str(path).lower().endswith('.h5'):
            return True
        with open(path, 'rb') as f:
            r = f.read(8) == bytes.fromhex('894844460d0a1a0a')
            return r


class ONNX(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = ModelFormat.ONNX

    @classmethod
    def is_me(cls, model: Model):
        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False
        if str(path).lower().endswith('.onnx'):
            return True

        import onnx
        try:
            onnx.checker.check_model(str(path))
        except onnx.onnx_cpp2py_export.checker.ValidationError:
            return False
        return True


class PTH(Identifier):
    """
    Handle model serialized by torch.save
    Pth archives the model in zip format.

    This identifier determin if the input file is in zip format by magic
    number, then get the file list and check if it match torch strucutre
    """
    FORMAT = ModelFormat.PTH

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        with open(path, 'rb') as f:
            if not f.read(4) == bytes.fromhex('504b0304'):
                return False

        try:
            z = zipfile.ZipFile(path)
            file_names = '|'.join(z.namelist())
            return ('code/__torch__' not in file_names) and \
                   'data.pkl' in file_names and \
                   'data/' in file_names
        except OSError:
            return False


class TorchTraced(Identifier):
    """
    Torchscript archives the model in zip format.

    This identifier determin if the input file is in zip format by magic
    number, then get the file list and check if it match torch strucutre
    """
    FORMAT = ModelFormat.TORCH_TRACED

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        with open(path, 'rb') as f:
            if not f.read(4) == bytes.fromhex('504b0304'):
                return False

        try:
            z = zipfile.ZipFile(path)
            file_names = '|'.join(z.namelist())
            return 'code/__torch__' in file_names and \
                   'data.pkl' in file_names and \
                   'data/' in file_names
        except OSError:
            return False


class PB(Identifier):
    """
    Use file extension and magic number to identify the file
    """

    FORMAT = ModelFormat.PB

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        # I dont find such pattern in:
        # 1. https://github.com/chen0040/java-tensorflow-samples/blob/master/audio-classifier/src/main/resources/tf_models/resnet-v2.pb
        # 2. https://github.com/bugra/putting-tensorflow-2-models-to-production/blob/master/models/resnet/1538687457/saved_model.pb
        # 3. https://github.com/U-t-k-a-r-s-h/Auto-Labeling-tool-using-Tensorflow/blob/master/Mobilenet.pb
        # 4. https://codechina.csdn.net/shy_201992/human-pose-estimation-opencv/-/blob/master/graph_opt.pb
        #
        # with open(path, 'rb') as f:
        #     if f.read(8) == 'PBDEMS2\0':
        #         return True

        # check if 'dtype' exists in the first 1k of the file.
        with open(path, 'rb') as f:
            if bytes.fromhex('0A05647479706512') in f.read(1024):
                return True

        return str(path).lower().endswith('.pb')


class TFLITE(Identifier):
    """
    Use file extension and magic number to identify the file
    """

    FORMAT = ModelFormat.TFLITE

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if str(path).lower().endswith(".tflite"):
            return True
        if not (isinstance(path, Path) and path.exists() and path.is_file()):
            return False

        with open(path, 'rb') as f:
            if bytes.fromhex('1C00000054464C33') in f.read(8):
                return True

        return False


class SavedModel(Identifier):
    """
    Use directory pattern to identify the file
    """

    FORMAT = ModelFormat.SAVED_MODEL

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists()):
            return False

        if path.is_dir():
            return (os.path.exists(os.path.join(path, 'saved_model.pb')) and
                    os.path.exists(os.path.join(path, 'variables')))
        return False


class ZippedSavedModel(Identifier):
    """
    Use directory pattern to identify the file
    """

    FORMAT = ModelFormat.ZIPPED_SAVED_MODEL

    @classmethod
    def is_me(cls, model: Model) -> bool:

        if model.format == cls.FORMAT:
            return True

        if isinstance(model.src, str):
            path = Path(model.src)
        else:
            path = model.src

        if not (isinstance(path, Path) and path.exists()):
            return False

        if path.is_file():
            try:
                z = zipfile.ZipFile(path)
                file_names = '|'.join(z.namelist())
                return 'saved_model.pb' in file_names and 'variables' in file_names
            except OSError:
                return False
            except zipfile.BadZipFile:
                return False
        return False


class TFKerasModel(Identifier):
    ''' Use python MRO to check if it contains specific str
    '''

    FORMAT = ModelFormat.TF_KERAS_MODEL

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        return _check_object_type('tensorflow.python.keras', model.src)


class TFSession(Identifier):
    ''' Use python MRO to check if it contains specific str
    '''
    FORMAT = ModelFormat.TF_SESSION

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        return _check_object_type('tensorflow.python.client.session.Session',
                                  model.src)


class KerasModel(Identifier):
    '''Use python MRO to check if it contains specific str
    Keras 2.5.0 Serializer
    '''

    FORMAT = ModelFormat.KERAS_MODEL

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        return _check_object_type('keras.', model.src)


class PytorchModel(Identifier):
    """Use python MRO to check if it contains specific str"""

    FORMAT = ModelFormat.PT_NN_MODULE

    @classmethod
    def is_me(cls, model: Model):

        if model.format == cls.FORMAT:
            return True

        if _check_object_type('torch.nn.module', model.src):
            return True

        elif _check_object_type('torchvision.models.', model.src):
            return True

        else:
            return False


class OpenvinoIRDir(Identifier):
    FORMAT = ModelFormat.OPENVINO_IRDIR

    @classmethod
    def is_me(cls, model: Model):
        model_path = model.src
        if not os.path.isdir(model_path):
            return False
        files = glob.glob(f"{model_path}/*")

        return (any([f.endswith(".xml") for f in files]) and
                any([f.endswith(".bin") for f in files]) and
                any([f.endswith(".mapping") for f in files]))


class TensorrtPLAN(Identifier):
    FORMAT = ModelFormat.TRT_PLAN

    @classmethod
    def is_me(cls, model: Model):
        model_path = model.src
        if not os.path.isfile(model_path):
            return False
        return str(model_path).endswith(".plan") or str(model_path).endswith(
            ".engine")



from typing import Union, Tuple, Any, List, Type, Dict, Optional
from dataclasses import dataclass
from pathlib import Path
from collections import OrderedDict
from loguru import logger
from enum import Enum

from . import ModelFormat, ModelDataType
from ..common import get_class_name


@dataclass
class Tensor:
    name: str
    shape: Union[None, Tuple[int, ...]] = None
    dtype: Union[None, ModelDataType] = None

    def dump(self) -> Dict:
        return {
            "name":
                self.name,
            "shape":
                self.shape,
            "type":
                self.dtype.name
                if isinstance(self.dtype, ModelDataType) else None
        }


class Model():

    def __init__(self,
                 src: Union[str, Path, object],
                 format=ModelFormat.NON_SPECIFIED,
                 inputs: List[Union[List, Tuple, Tensor]] = None,
                 outputs: List[Union[List, Tuple, Tensor]] = None,
                 batch_dim=0):
        """
        Do not put identify_format in Model. Make sure
        (MVC) control and model are separated.
        """
        if not inputs:
            inputs = []
        if not outputs:
            outputs = []
        if isinstance(src,str):
            self.src = Path(src)
        else:
            self.src = src
        self.inputs: List[Tensor] = inputs
        self.outputs: List = outputs
        self.format = format
        self.batch_dim: int = batch_dim

        self._name: str
        if isinstance(src, str):
            self._name = Path(src).stem
        elif isinstance(src, Path):
            self._name = src.stem
        else:
            self._name = str(type(object))

    @property
    def name(self) -> str:
        return self._name

    def set_name(self, _name: str):
        self._name = _name

    def set_batch_dim(self, _batch_dim: Optional[int]):
        if _batch_dim is not None:
            assert _batch_dim >= 0
        self.batch_dim = _batch_dim

    def reset_inputs(self, input_: List[Tensor]) -> None:
        self.inputs = []
        for tensor in input_:
            self.inputs.append(tensor)

    def reset_outputs(self, output_: List[Tensor]) -> None:
        self.outputs = []
        for tensor in output_:
            self.outputs.append(tensor)

    def clone_attributes(self, model: Any) -> None:
        """Clone user defined attributes from input

        Iter all user-defined attributes and check if they have values.
        If the attribute of the destination object has value, then skip
        the clone.

        Args:
            model Type[Dataset]: The model object to be cloned
        """
        _to_be_cloned: List = []
        for attr_name in model.__dict__:

            dst_value = getattr(self, attr_name)
            src_value = getattr(model, attr_name)

            if (isinstance(dst_value, OrderedDict)) or (isinstance(
                    dst_value, List)):
                if len(dst_value) == 0:
                    _to_be_cloned.append([attr_name, src_value])

            elif dst_value is None:
                _to_be_cloned.append([attr_name, src_value])

        for attr_name, src_value in _to_be_cloned:
            setattr(self, attr_name, src_value)

    def dump(self) -> Dict:
        """
        WARNING: the dumped dictionary can only be used for report
        since src can be path or class name (which is not a good design)
        """
        data = {
            "src": "",
            "name": self.name,
            "inputs": {x.name: x.dump() for x in self.inputs},
            "outputs": {x.name: x.dump() for x in self.outputs},
            "format": self.format.name
        }
        if isinstance(self.src, str):
            data["src"] = self.src
        elif isinstance(self.src, Path):
            data["src"] = str(self.src)
        else:
            data["src"] = get_class_name(self.src)

        return data

    def load(self, obj: Dict):
        raise NotImplementedError()



from .core.options import Options

# Options initialize is required for OptionsRegistry to work properly
from onnc.forest.runtimes.openvino.runtime import OpenvinoOptions
from onnc.forest.runtimes.onnx.runtime import ONNXOptions
from onnc.forest.runtimes.tensorrt.runtime import TensorRTOptions



from __future__ import annotations
import os
import zipfile
from typing import List
from pathlib import Path

import numpy as np
import onnx
from loguru import logger

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dataset import Dataset

from .dataset import DatasetFormat


class IdentifierRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Identifier(metaclass=IdentifierRegistry):

    FORMAT: DatasetFormat = DatasetFormat.NON_SPECIFIED

    @classmethod
    def is_me(self, dataset: Dataset) -> bool:
        pass


def identify(dataset: Dataset) -> Dataset:
    for identifier in Identifier.REGISTRY:
        if identifier.is_me(dataset):
            return identifier.FORMAT

    raise NotImplementedError(f"Unable to identify {dataset.src}")


class NONE(Identifier):
    FORMAT = DatasetFormat.NONE

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return dataset.src is None

class PB(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.PB

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        if isinstance(dataset.src, str) or isinstance(dataset.src, Path):
            path = Path(dataset.src)
        else:
            return False

        if not (path.exists() and path.is_file()):
            return False

        if str(path).lower().endswith('.pb'):
            return True

        try:
            tensor = onnx.TensorProto()
            with open(dataset.src, 'rb') as f:
                tensor.ParseFromString(f.read())
            dims = list(tensor.dims)
        except:
            return False

        return True if dims else False


class NPY(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPY

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        if isinstance(dataset.src, str) or isinstance(dataset.src, Path):
            path = Path(dataset.src)
        else:
            return False

        if not (path.exists() and path.is_file()):
            return False
        if str(path).lower().endswith('.npy'):
            return True

        with open(path, 'rb') as f:
            return f.read(6) == bytes.fromhex('934E554D5059')


class NPYDIR(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPYDIR

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        if isinstance(dataset.src, str) or isinstance(dataset.src, Path):
            path = Path(dataset.src)
        else:
            return False

        if not (path.exists() and path.is_dir()):
            return False
        files = os.listdir(path)
        if not files:
            return False
        return str(files[0]).lower().endswith('.npy')


class NPZ(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPZ

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        if isinstance(dataset.src, str) or isinstance(dataset.src, Path):
            path = Path(dataset.src)
        else:
            return False

        if not (path.exists() and path.is_file()):
            return False

        if str(path).lower().endswith('.npy'):
            return True

        try:
            with zipfile.ZipFile(str(path)) as archive:
                for name in archive.namelist():
                    if not name.endswith('.npy'):
                        continue

                    npy = archive.open(name)
                    version = np.lib.format.read_magic(npy)
                    return True
        except Exception as e:
            logger.debug(str(e))

        return False


class NP_ndarray(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NDARRAY

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return type(dataset.src) is np.ndarray


class NP_memmap(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NP_MEMMAP

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return type(dataset.src) is np.memmap


class NPZFileObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.NPZ_OBJECT

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        return dataset.src.__class__.__name__ == "NpzFile"


class TorchDatasetObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TORCH_DATASET

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        # from torch.utils.data.dataset import Dataset as TorchDataset
        # return isinstance(dataset.src, TorchDataset)
        str_type = str(dataset.src.__class__.__mro__)
        return '.Dataset' in str_type and 'torch.' in str_type


class TorchDataloaderObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TORCH_DATALOADER

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        # from torch.utils.data.dataloader import DataLoader
        # return isinstance(dataset.src, DataLoader)
        str_type = str(dataset.src.__class__.__mro__)
        return '.DataLoader' in str_type and 'torch.' in str_type


class KerasDatasetObject(Identifier):
    """
    Keras Predefined dataset
    https://www.tensorflow.org/api_docs/python/tf/keras/datasets
    """
    FORMAT = DatasetFormat.KERAS_DATASET

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True

        ds = dataset.src
        # Keras dataset is a 2x2 metric of tuple
        # ds[0] is for training , ds[0][0] is x and ds[0][1] is y
        # ds[0] is for testing  , ds[1][0] is x and ds[1][1] is y
        # each record is a np.ndarray object
        if isinstance(ds, (tuple, list)):
            if len(ds) == 2 and len(ds[0]) == 2 and len(ds[1]) == 2:
                if all([
                        isinstance(x, np.ndarray)
                        for x in [ds[0][0], ds[0][1], ds[1][0], ds[1][1]]
                ]):
                    return True
        return False


class TFDSPrefetchObject(Identifier):
    """
    Use file extension and magic number to identify the file
    """
    FORMAT = DatasetFormat.TFDS_PREFETCH

    @classmethod
    def is_me(cls, dataset):
        if dataset.format == cls.FORMAT:
            return True
        str_type = str(type(dataset.src))
        return '.PrefetchDataset' in str_type and 'tensorflow.' in str_type



from __future__ import annotations
from typing import Callable, Dict, List, Type
import types

from .identifier import identify, DatasetFormat
from .transform import DatasetTransformer
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dataset import Dataset


class PreprocessTransformer(DatasetTransformer):

    FORMAT = DatasetFormat.NON_SPECIFIED

    @classmethod
    def is_me(cls, dataset: Dataset) -> bool:
        return identify(dataset) == cls.FORMAT

    def transform(self, dataset: Type[Dataset]):
        pass

    def set_preprocessor(self, func: Callable, args: List = [],
                         kwargs: Dict = {}) -> None:

        """Set the transform func as given callback

        Args:
            func (Callable): preprocessor function, the output of the this
                             function has to be the preprocessed dataset.
            args (List): args for the preprocessor function, use `__DATASET__`
                         to indicate the dataset argument position

            kwargs (Dict): keword args for the preprocessor function

        Example:

            def _preprocess(dataset, min_val, max_val, enhancement=False):
              ....
              ....

            set_preprocessor(_preprocess,
                            ['__DATASET__', 1, 100],
                            {'enhancement': True}
                            )
        """

        self.set_transform(func, args, kwargs)



from abc import ABC, abstractmethod
from onnc.bench.core.model.identifier import identify
from onnc.bench.core.model.model import Model
from onnc.forest.core.options import Options

from typing import Union, List


class Runtime:
    """
    This class works like a wrapper which is used for:
    - hidding runtime factory from user
    - narrow down runtime lifecycle (not init until load)
    """

    def __init__(self, options=None):
        self._options = options
        self.loaded_runtimes = []

        self.__init__state: List[bool] = [False, False, False, False, False]
        # indicate whether `launch`, `load`, `bind_input`, `bind_output`,
        #  `materialize` are trigger.

    def launch(self, options):
        self.__init__state[0] = True

    def load(self, options):
        self.__init__state[1] = True
        self.loaded_runtimes.append(make_runtime(options))
        idx = len(self.loaded_runtimes) - 1

        self.loaded_runtimes[idx].launch(options)
        self.loaded_runtimes[idx].load(options)

    def bind_input(self, input: Union[int, str], object: object):
        self.__init__state[2] = True
        return self.loaded_runtimes[0].bind_input(input, object)

    def bind_output(self, output: Union[int, str], object: object):
        self.__init__state[3] = True
        return self.loaded_runtimes[0].bind_output(output, object)

    def bind_all_inputs(self):
        self.__init__state[2] = True
        return self.loaded_runtimes[0].bind_all_inputs()

    def bind_all_outputs(self):
        self.__init__state[3] = True
        return self.loaded_runtimes[0].bind_all_outputs()

    def materialize(self):
        self.__init__state[4] = True
        res = []

        for rt in self.loaded_runtimes:
            res.append(rt.materialize())

        return res

    def read(self, *args):
        return self.loaded_runtimes[0].read(*args)

    def run(self):
        return self.loaded_runtimes[0].run()

    def write(self, *args):
        return self.loaded_runtimes[0].write(*args)

    def __call__(self, inputs: List):
        if not all(self.__init__state):
            self.launch(self._options) if self.__init__state[0] == 0 else None
            self.load(self._options) if self.__init__state[1] == 0 else None
            self.bind_all_inputs() if self.__init__state[2] == 0 else None
            self.bind_all_outputs() if self.__init__state[3] == 0 else None
            self.materialize() if self.__init__state[4] == 0 else None

        self.write(inputs)
        self.run()
        return self.read()


class AbstractAsyncRuntimeImplementation(ABC):

    def __init__(self):
        pass

    @abstractmethod
    def launch(self, options):
        pass

    @abstractmethod
    def load(self, options):
        pass

    @abstractmethod
    def bind_input(self):
        pass

    @abstractmethod
    def bind_output(self):
        pass

    @abstractmethod
    def materialize(self):
        pass

    @abstractmethod
    def submit(self, data):
        pass

    @abstractmethod
    def wait(self):
        pass

    @abstractmethod
    def outcome(self):
        pass


class RuntimeRegistry(type):

    REGISTRY = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class AbstractRuntimeImplementation(metaclass=RuntimeRegistry):

    @staticmethod
    def is_me(format) -> bool:
        return False

    def __init__(self):
        pass

    @abstractmethod
    def launch(self, options):
        pass

    @abstractmethod
    def load(self, options):
        pass

    @abstractmethod
    def bind_input(self):
        pass

    @abstractmethod
    def bind_output(self):
        pass

    @abstractmethod
    def materialize(self):
        pass

    @abstractmethod
    def write(self):
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def read(self, data):
        pass


def make_runtime(options: Options) -> AbstractRuntimeImplementation:

    for runtime in RuntimeRegistry.REGISTRY:
        if runtime.is_me(options):
            return runtime()
    raise NotImplementedError(f"Unable to identify {options}")



from typing import Dict


class DataLayout:
    def __init__(self, batch_axis: int):
        self.batch_axis = batch_axis

    def add_axis(self, name: str, axis: int) -> None:
        setattr(self, name, axis)

    def dump(self) -> Dict[str, int]:
        data = self.__dict__
        del data["dump"]
        del data["add_axis"]
        return data


class ImageDataLayout(DataLayout):
    def __init__(self, batch_axis: int, channel_axis: int,
                 hight_axis: int, width_axis: int) -> None:
        self.batch_axis = batch_axis
        self.channel_axis = channel_axis
        self.hight_axis = hight_axis
        self.width_axis = width_axis



from pathlib import Path
from typing import Dict, Union, List
from pathlib import Path
import os
import json
import shutil

from loguru import logger
from onnc.bench.core.deployment import Deployment
from .builder import IBuilder
from onnc.bench.core.common import get_tmp_path
from . import Compilation
from onnc.bench.core.model.model import Model
from onnc.bench.core.dataset.dataset import Dataset


def spinning_cursor():
    while True:
        for cursor in '|/-\\':
            yield cursor


class NNUXEBuilder(IBuilder):
    BUILDER_NAME = "NNUXEBuilder"

    def __init__(self):
        self._compilations: Dict[int, Compilation] = {}
        self.output_path: str = ""


    def _compile(self,
                 model_name,
                 model_path: str,
                 sample_path: str,
                 params_path: str,
                 output_path: str,
                 local_nnuxe: bool = True):

        from nnuxe.drivers.compiler import compile as nnuxe_compile
        from nnuxe.core.report import CompileReport
        report = CompileReport()
        nnuxe_compile(model_path,
                      sample_path,
                      params_path,
                      os.path.join(output_path, model_name),
                      report,
                      local_nnuxe=local_nnuxe)
        return report

    def build(self, target: str, converter_params={}) -> Dict:

        compilation_list = []
        params = {}
        res = {}
        # Upload files and create compilation
        for iternal_cid in self._compilations:
            compilation = self._compilations[iternal_cid]

            compilation.compiler_params['target'] = target

            params["model_meta"] = compilation.model_meta
            params["sample_meta"] = compilation.sample_meta
            params["compiler_params"] = compilation.compiler_params
            params["calibrator_params"] = compilation.calibrator_params
            params["converter_params"] = converter_params
            compilation_list.append(
                (compilation.model_path, compilation.sample_path, params))
        output_path = get_tmp_path()
        os.makedirs(output_path, exist_ok=True)
        for idx, compi in enumerate(compilation_list):
            model_path = compi[0]
            sample_path = compi[1]
            params = compi[2]
            params_path = get_tmp_path()
            open(params_path, 'w').write(json.dumps(params))
            report = self._compile(f'model_{idx}', model_path, sample_path,
                                   params_path, output_path)
            res[f'model_{idx}'] = report
            report.dump_json(
                os.path.join(output_path, f'model_{idx}', "report.json"))

            os.remove(params_path)

            logger.debug(params)

        self.output_path = output_path

        return res

    def save(self, output: Path) -> Union[Dict, Deployment]:
        res = {}
        shutil.rmtree(output, ignore_errors=True)
        shutil.copytree(self.output_path, output)
        shutil.rmtree(self.output_path, ignore_errors=True)

        for i in os.listdir(output):
            try:
                out = Path(os.path.join(output, i))
                deployment = Deployment(out)
            except Exception as e:
                deployment = Deployment(None)

            res[i] = deployment
        return res
        # return Deployment(None, report=res, logs=res)

    @property
    def supported_devices(self) -> Dict:
        devices = {
            'CMSIS-NN': 'CMSIS-NN',
            'ANDES-LIBNN': 'ANDES-LIBNN',
            'NVDLA-NV-SMALL': 'NVDLA-NV-SMALL',
            'NVDLA-NV-LARGE': 'NVDLA-NV-LARGE',
            'NVDLA-NV-FULL': 'NVDLA-NV-FULL',
            'CMSIS-NN-DEFAULT': 'CMSIS-NN-DEFAULT',
            'NVDLA-NV-SMALL-DEFAULT': 'NVDLA-NV-SMALL-DEFAULT',
            'NVDLA-NV-LARGE-DEFAULT': 'NVDLA-NV-LARGE-DEFAULT',
            'NVDLA-NV-FULL-DEFAULT': 'NVDLA-NV-FULL-DEFAULT',
            'NVIDIA-TENSORRT-FP32' : 'NVIDIA-TENSORRT-FP32',
            'NVIDIA-TENSORRT-FP16' : 'NVIDIA-TENSORRT-FP16',
            'NVIDIA-TENSORRT-INT8' : 'NVIDIA-TENSORRT-INT8',
            'RELAYIR' : 'RELAYIR',
            "INTEL-OPENVINO-CPU-FP32": "INTEL-OPENVINO-CPU-FP32",
            "FIXED_ONNX": "FIXED_ONNX",
            "ONNX": "ONNX",
            "ONNC-IN2O3": "ONNC-IN2O3",
            "GenericONNC": "GenericONNC"
        }
        return devices



import tempfile
import threading
import os


def get_class_name(model):
    c = model.__class__.__mro__[0]
    name = c.__module__ + "." + c.__name__
    return name


def get_temp_base():
    tid = str(threading.get_ident())
    return os.path.join(tempfile._get_default_tempdir(), 'onnc-bench', tid)


def get_tmp_path():
    temp_base = get_temp_base()

    if not os.path.exists(temp_base):
        os.makedirs(temp_base, exist_ok=True)
    return os.path.join(temp_base, next(tempfile._get_candidate_names()))



import os
import pkg_resources
from packaging import version
from loguru import logger

protobuf_ver = pkg_resources.get_distribution("protobuf").version
if version.parse(protobuf_ver) > version.parse('3.20.3'):
    if not os.environ.get('PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION'):
        os.environ['PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION'] = "python"
        logger.warning(
            "Set environment variable: "
            "PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION=python. Reason: protobuf "
            f"version {protobuf_ver} is found in your system, however, onnx "
            "only supports protobuf package == 3.20.x or lower. The "
            "environment variable makes onnx use pure-Python to parse protobuf "
            "format and will be much slower. Alternatively, you can downgrade "
            "the protobuf package to 3.20.x or lower.)")



import os
from os.path import expanduser
import requests
from typing import Union

from .core.compiler.onnc_saas import ONNCSaaSBuilder

from .core.compiler.builder import IBuilder
try:
    if int(os.environ['ONNC_TESTMODE']):
        api_host = 'http://127.0.0.1:8000'
except KeyError:
    api_host = 'https://api.onnc.skymizer.com'

onnc_key_var = 'ONNC_APIKEY'


api_protocol = 'https'
api_url = "api.onnc.skymizer.com"
api_port = 443

image_name = 'registry.skymizer.com/nnuxe/nnuxe/nnuxe_image'
image_tag = "v0.0.4-test"

default_builder: IBuilder
try:
    if os.environ['BENCH_COMPILER'] == "NNUXE":
        from .core.compiler.nnuxe import NNUXEBuilder

        default_builder = NNUXEBuilder
        default_builder_params = []
    elif os.environ['BENCH_COMPILER'] == "NNUXE_DOCKER":
        from .core.compiler.nnuxe_docker import NNUXEDockerBuilder

        default_builder = NNUXEDockerBuilder
        default_builder_params = [image_name, image_tag]
    else:
        default_builder = ONNCSaaSBuilder
        default_builder_params = [api_protocol, api_url, api_port]
except KeyError:
    default_builder = ONNCSaaSBuilder
    default_builder_params = [api_protocol, api_url, api_port]



import msgpack
import msgpack_numpy as m
import numpy as np

def np_to_bytes(payload: np.array)->bytes:
    """
    to_bytes takes a numpy.array as a single argument and serializes it to bytes.
    """
    return msgpack.packb(payload, default=m.encode)


def np_from_bytes(payload:bytes) -> np.array:
    """
    from_bytes recovers a numpy.array from a bytes array that was produced via numpy_serializer.to_bytes.
    """ 
    return msgpack.unpackb(payload, object_hook=m.decode)


from pathlib import Path
from typing import Dict, Union, List, Tuple, Optional
import os
import json
import shutil

from loguru import logger

from .core.modelpackage import ModelPackage
from .core.dataset.dataset import Dataset

from .core.model import ModelFormat, TF_MODEL_FORMATS, ModelDataType
from .core.model.model import Tensor

from .core.dataset import DatasetFormat

from .core.compiler.onnc_saas import ONNCSaaSBuilder
# from .core.compiler.nnuxe import NNUXEBuilder
# from .config import api_protocol, api_url, api_port
from .core.common import get_tmp_path, get_temp_base

from .config import default_builder, default_builder_params
from onnc.bench.core.model.model import Model


def login(api_key: str, password: str = ""):
    os.environ['ONNC_API_USER'] = api_key
    os.environ['ONNC_API_PASSWD'] = password


def get_api_key():
    return (os.environ.get('ONNC_API_USER'), os.environ.get('ONNC_API_PASSWD'))


class Project:

    def __init__(self,
                 name,
                 builder=default_builder,
                 builder_params=default_builder_params):
        self.builder = builder(*builder_params)

        if isinstance(self.builder, ONNCSaaSBuilder):
            self.builder.saas_login(*get_api_key())
            proj_info_sr = self.builder.saas_get_project_id_by_name(name)
            if len(proj_info_sr.data) == 0:
                self.builder.saas_create_project(name)

            logger.info(f"ProjectID: {self.builder._project_id}")

    def add_model(
            self,
            model: Union[object, str, Path],
            samples: Union[object, str, Path, Dataset, None] = None,
            model_format: ModelFormat = ModelFormat.NON_SPECIFIED,
            sample_format: DatasetFormat = DatasetFormat.NON_SPECIFIED,
            # [name: str, shape:Tuple, dtype:ModelDataType]
            model_inputs: List[Union[List, Tuple, Tensor]] = [],
            # [name: str, shape:Tuple, dtype:ModelDataType]
            model_outputs: List[Union[List, Tuple, Tensor]] = [],
            inputs_as_nchw: Union[str, bool, None, List[str]] = None,
            batch_dim: Optional[int] = 0):
        """Add a model and its corresponding calibration samples.

        :param Union[object, str, Path] model:
            A file or a directory path to a serialized model file,
            or a neurl network model object.

            Check https://docs-tinyonnc.skymizer.com/model-formats.html for
            a list of supported formats.

        :param Union[object, str, Path, Dataset] samples:
            A file or a directory path to a serialized dataset or object.

            Check https://docs-tinyonnc.skymizer.com/dataset-formats.html for
            a list of supported formats.

        :param ModelFormat model_format:
            The format/framework of the model file/object. onnc-bench will
            identify the format if this param is ModelFormat.NON_SPECIFIED.

            Check https://docs-tinyonnc.skymizer.com/model-formats.html for
            a list ofsupported formats.

        :param DatasetFormat sample_format:
            The format/framework of the model file/object. onnc-bench will
            identify the format if this param is DatasetFormat.NON_SPECIFIED.

            Check https://docs-tinyonnc.skymizer.com/dataset-formats.html for
            a list of supported formats.
        :param List[Union[List, Tuple, Tensor]] model_inputs:
            A list of input tensor(s), which include(s) 3 fields:
                1. name: str
                2. shape(optional): Union[List, Tuple]
                3. dtype(optional): onnc.bench.core.model.ModelDataType

        :param List[Union[List, Tuple, Tensor]] model_outputs:
            A list of input tensor(s), which include(s) 3 fields:
                1. name: str
                2. shape(optional): Union[List, Tuple]
                3. dtype(optional): onnc.bench.core.model.ModelDataType


        """
        model_package = ModelPackage(model, samples, model_format,
                                     sample_format, model_inputs, model_outputs,
                                     batch_dim)
        serialized_model, serialized_dataset = model_package.serialize()

        model_id = self.builder.prepare_model(serialized_model,
                                              serialized_dataset,
                                              inputs_as_nchw)
        self.builder.set_compilations_params(model_id,compile_params={},
            calibrator_params={})

    def compile(self, target: str, converter_params: Dict = None):
        """Trigger the compilation process and transform the given
        model into C function calls.

        :param str device:
            The name of the supported SoC board, for example, NUMAKER_IOT_M487.

            Check https://docs-tinyonnc.skymizer.com/devices.html for
            a list of supported devices.

        : param dict params:
            Parameters for the compiler.

            Check https://docs-tinyonnc.skymizer.com/compiler-params.html for
            more details.

        """
        if converter_params is None:
            converter_params = {}
        target_id = self.builder.get_device_id(target)
        res = self.builder.build(target_id, converter_params)

        if isinstance(self.builder, ONNCSaaSBuilder):
            logger.info(
                f"You review compilation report at https://app.onnc.skymizer.com"
                f"/app/builds/{res['id']}")

        return res

    def save(self, path: Union[str, Path]):
        """Save the compiled artifact(s).

        :param Union[str, Path] path:
            A path to store the artifact(s).

        :returns:
            A Deployment object that contains `files`, `report`.
        :rtype:
            Deployment
        """
        if isinstance(path, str):
            path = Path(path)
        return self.builder.save(path)

    def prune(self, params={}):
        raise NotImplementedError("Model pruning is not implemented yet.")

    """
    This require fix? Always raise error while compile
    """
    # def __del__(self):
    #     shutil.rmtree(get_temp_base())



import sys

from loguru import logger
import sentry_sdk

from .config import *


from .project import Project, login
from .core.compiler.onnc_saas import ONNCSaaSBuilder
from .config import api_protocol, api_url, api_port
from .core.deployment import Deployment
# from .workspace import *
# from .workspacev2 import CompressFormt, ModelFormat, SampleFormat, SampleSource
# from .workspacev2 import Samples, launch

try:
	logger.remove(0)
except:
	pass

logger.start(sys.stdout, colorize=True, format="<level>{level}</level> {message}", level="INFO")

sentry_sdk.init(
    "https://da9cdf5759874504940714a91657de21@o304393.ingest.sentry.io/5901378",
    traces_sample_rate=1.0
)
fpath = os.path.dirname(os.path.abspath(__file__))

__version__= "4.1.4"




from typing import Any, List, Union, Dict
from abc import abstractmethod
from pathlib import Path

from loguru import logger

from onnc.bench.core.deployment import Deployment
from onnc.bench.core.model.model import Model
from onnc.bench.core.dataset.dataset import Dataset
from onnc.bench.core.compiler import Compilation


class IBuilder:

    BUILDER_NAME = ""

    def __init__(self):
        self.model_ids: List = []

    def prepare_model(
            self,
            model: Model,
            dataset: Dataset,
            inputs_as_nchw: Union[str, bool, None, List[str]] = None) -> int:
        """ Upload a model and its corresponding calibration dataset.
            And create a compilation.

        :param str model:
            A path to a model file
        :param str dataset:
            A path to a model file
        :rtype:
            int
        """

        compilation = Compilation(model, dataset, inputs_as_nchw)
        _internal_cid = id(model.src)
        self._compilations[_internal_cid] = compilation
        return _internal_cid

    def set_compilations_params(self, model_id: int, compile_params: Dict,
        calibrator_params:Dict):
        self._compilations[model_id].calibrator_params = calibrator_params
        self._compilations[model_id].compiler_params = compile_params

    @abstractmethod
    def build(self, target, converter_params: Dict = {}) -> Any:
        """build a project witch contains multiple models

        for model in model_ids:
            ...
        """
        pass

    @abstractmethod
    def save(self, output: Path) -> Union[Dict, Deployment]:
        pass

    @property
    def supported_devices(self) -> Dict:
        pass

    def get_device_id(self, target):
        if target in self.supported_devices:
            return target
        else:
            logger.error(f'`{target}` is not a supported device/format.')
            logger.error(f'Supported devices/formats are: '
                         f'{str([x for x in self.supported_devices])} ')
            raise Exception(f'`{target}` is not supported')



from dataclasses import dataclass, field
from typing import List
import numpy as np
from importlib import import_module

from onnc.forest.core.runtime import AbstractRuntimeImplementation
from onnc.forest.core.options import Options, RuntimeSettings
from onnc.bench.core.model import ModelFormat


@dataclass
class TensorRTSettings(RuntimeSettings):
    pass


class TensorRTOptions(Options):
    model_format = ModelFormat.TRT_PLAN

    def __init__(self, loadable):
        self.settings: TensorRTSettings = TensorRTSettings()
        self.devices: List = []
        self.loadable = loadable


class TensorRTRuntime(AbstractRuntimeImplementation):

    @staticmethod
    def is_me(options: Options) -> bool:
        return isinstance(options, TensorRTOptions)

    def __init__(self):
        self.__pycuda = import_module('pycuda.driver')
        """"
        WARNING!!! DO NOT REMOVE pycuda.autoinit even if this module
        is not written in the code.
        """
        import_module('pycuda.autoinit')
        self._context = None
        self._inputs: List = []
        self._outputs: List = []
        self._bindings: List = []

    def launch(self, options):
        pass

    def load(self, options):
        import tensorrt as trt

        model_path = str(options.loadable.src)
        TRT_LOGGER = trt.Logger(trt.Logger.INFO)
        with open(model_path, "rb") as f, \
            trt.Runtime(TRT_LOGGER) as runtime:
            self._cuda_engine = runtime.deserialize_cuda_engine(f.read())

    def bind_input(self, input, object):
        # object: and HostDeviceMem object or tuple(host_mem, device_mem)
        from onnc.forest.runtimes.tensorrt.utils import HostDeviceMem

        if isinstance(object, HostDeviceMem):
            self._inputs.append(object)
            self._bindings.append(object.device)
        else:
            assert type(object) in [tuple, list]
            assert len(object) == 2
            assert type(object[0]) is int and type(object[1]) is int
            self._inputs.append(HostDeviceMem(*object))
            self._bindings.append(object[1])

    def bind_output(self, output, object):
        # object: and HostDeviceMem object or tuple(host_mem, device_mem)
        from onnc.forest.runtimes.tensorrt.utils import HostDeviceMem

        if isinstance(object, HostDeviceMem):
            self._outputs.append(object)
            self._bindings.append(object.device)
        else:
            assert type(object) in [tuple, list]
            assert len(object) == 2
            assert type(object[0]) is int and type(object[1]) is int
            self._outputs.append(HostDeviceMem(*object))
            self._bindings.append(object[1])

    def bind_all_inputs(self):
        from onnc.forest.runtimes.tensorrt.utils import allocate_buffers

        self._inputs, self._outputs, self._bindings, _ = \
            allocate_buffers(self._cuda_engine)

    def bind_all_outputs(self):
        pass

    def materialize(self):
        self._context = self._cuda_engine.create_execution_context()

    def write(self, data):
        for i, d in enumerate(data):
            np.copyto(self._inputs[i].host, d.ravel())
        for inp in self._inputs:
            self.__pycuda.memcpy_htod(inp.device, inp.host)

    def run(self):
        if self._context is None:
            raise RuntimeError("Please run materialize before running infer")

        self._context.execute_v2(bindings=self._bindings)

    def read(self):
        for out in self._outputs:
            self.__pycuda.memcpy_dtoh(out.host, out.device)

        return [out.host for out in self._outputs]



timeout = 300  # in seconds
poll_interval = 0.5  # in seconds

URI_MAP = {
    "saas_login": {
        "uri": "/users/login",
        "method": "POST"
    },
    "saas_create_project": {
        "uri": "/projects",
        "method": "POST"
    },
    "saas_get_project_id_by_name": {
        "uri": "/projects?filter[where][name]={}",
        "method": "GET"
    },
    "saas_upload_file": {
        "uri": "/files",
        "method": "POST"
    },
    "saas_create_compilation": {
        "uri": "/compilations",
        "method": "POST"
    },
    "saas_create_build": {
        "uri": "/builds",
        "method": "POST"
    },
    "saas_get_build_state": {
        "uri": "/builds/{}",
        "method": "GET"
    },
    "saas_download_deployment_package": {
        "uri": "/files/d/{}",
        "method": "GET"
    },
    "saas_upload_report": {
        "uri": "/compilation-reports",
        "method": "POST"
    },
    "saas_list_devices": {
        "uri": "/boards",
        "method": "GET"
    }

}



import sys
import pycuda.driver as cuda
import pycuda.autoinit
import time
import tensorrt as trt


class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


def allocate_buffers(engine):
    """Allocates host and device buffer for TRT engine inference.
    This function is similair to the one in common.py, but
    converts network outputs (which are np.float32) appropriately
    before writing them to Python buffer. This is needed, since
    TensorRT plugins doesn't support output type description, and
    in our particular case, we use NMS plugin as network output.
    Args:
        engine (trt.ICudaEngine): TensorRT engine
    Returns:
        inputs [HostDeviceMem]: engine input memory
        outputs [HostDeviceMem]: engine output memory
        bindings [int]: buffer to device bindings
        stream (cuda.Stream): cuda stream for engine inference synchronization
    """
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()

    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        """
         pagelocked means memory won't be swap with the secondary memory
         such as harddrive, but here we only use pagelocked_empty to
         calculate the size to allocate
        """
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream

# This function is generalized for multiple inputs/outputs.
# inputs and outputs are expected to be lists of HostDeviceMem objects.
def do_inference(context, bindings, inputs, outputs, stream,use_sync=False):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]

    # Run inference.
    if use_sync:
        context.execute_v2(
                bindings=bindings)
    else:
        context.execute_async_v2(
            bindings=bindings,
            stream_handle=stream.handle)

    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Blocks until stream has completed all operations
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]



from typing import Dict, List, Union, Type, Iterable, Any, Tuple, Optional
from pathlib import Path

from loguru import logger

from .dataset.dataset import Dataset
from .dataset.transformer import DatasetTransformer
from .evaluator import Evaluator
from .evaluator import Metric

from .dataset import serializer as dataset_serializer
from .dataset import DatasetFormat
from .dataset import identifier as dataset_identifier
from .dataset.dataset import Dataset

from .model import serializer as model_serializer
from .model import ModelFormat, ModelDataType
from .model import PYTORCH_MODEL_FORMATS
from .model import identifier as model_identifier
from .model.model import Model, Tensor
from .model.meta import retrieve_model_metadata
from .model.transformer import ModelTransformer

from .common import get_tmp_path


class ModelPackage:
    """Contain all data and controllers

    Data include: model and dataset
    Controllers include: evaluators, model_transformer and dataset_transformer

    model: Model CANNOT be type MODEL otherwise we won't know the priority
           of model.format and model_format
    dataset: same concept as model
    """

    def __init__(
            self,
            model: Union[str, Path],
            dataset: Union[str, Path] = None,
            model_format: ModelFormat = ModelFormat.NON_SPECIFIED,
            sample_format: DatasetFormat = DatasetFormat.NON_SPECIFIED,
            #  [name: str, shape:Tuple, dtype:ModelDataType]
            model_inputs: List[Union[List, Tuple, Tensor]] = None,
            #  [name: str, shape:Tuple, dtype:ModelDataType]
            model_outputs: List[Union[List, Tuple, Tensor]] = None,
            batch_dim: Optional[int] = 0):

        model_inputs = model_inputs if model_inputs else []
        model_outputs = model_outputs if model_outputs else []

        self.evaluators: List[Evaluator] = []
        self.model_transformers: List[ModelTransformer] = []
        self.dataset_transformers: List[DatasetTransformer] = []
        self.model = self.__encapsulate_model(model, model_format, model_inputs,
                                              model_outputs, batch_dim)
        self.dataset = self.__encapsulate_dataset(dataset, format=sample_format)

    def __encapsulate_model(self, model: Union[str, Path], model_format,
                            model_inputs, model_outputs, batch_dim) -> Model:
        """
        Do not put these code in Model.__init__(). Make sure
        (MVC) control and model are separated.
        """

        def inputs_outputs_to_tensors(data: List[Union[List, Tuple, Tensor]]):
            res = []
            for i in data:
                if isinstance(i, list) or isinstance(i, tuple):
                    assert len(i) > 0
                    name = i[0]
                    shape = i[1] if len(i) >= 2 else tuple()
                    dtype = i[2] if len(i) >= 3 else ModelDataType.NON_SPECIFIED
                    tensor = Tensor(name, shape, dtype)
                elif isinstance(i, Tensor):
                    tensor = i
                res.append(tensor)
            return res

        res_model = Model(model, batch_dim=batch_dim)
        if model_format == ModelFormat.NON_SPECIFIED:
            res_model.format = model_identifier.identify(res_model)
        else:
            res_model.format = model_format

        meta = retrieve_model_metadata(res_model)
        meta_input_names = set([x.name for x in meta.inputs])
        meta_output_names = set([x.name for x in meta.outputs])
        if len(model_inputs) == 0:
            if res_model.format in PYTORCH_MODEL_FORMATS:
                logger.error("Please specify model_inputs in "
                             "Project.add(model, model_inputs="
                             "[[input_name, shape, type]])."
                             "For example, Project.add(model, "
                             "model_inputs=[['input_1', [1, 3, 224, 224], "
                             "float]])")
                raise Exception(
                    "Parameter `model_inputs` is required for Pytorch formats")
            res_model.reset_inputs(meta.inputs)
        else:
            tensors = inputs_outputs_to_tensors(model_inputs)
            for tensor in tensors:
                if tensor.name not in meta_input_names:
                    logger.warning(f"It seems input tenosr({tensor.name}) "
                                   f"does not match with model input schema "
                                   f"({meta_input_names}).")
            res_model.reset_inputs(tensors)

        if len(model_outputs) == 0:
            res_model.reset_outputs(meta.outputs)
        else:
            tensors = inputs_outputs_to_tensors(model_outputs)
            for tensor in tensors:
                if tensor.name not in meta_output_names:
                    logger.warning(f"It seems output tenosr({tensor.name}) "
                                   f"does not match with model output schema "
                                   f"({meta_output_names}).")
            res_model.reset_outputs(tensors)
        return res_model

    def __encapsulate_dataset(self, dataset: Union[str, Path],
                              format: DatasetFormat) -> Dataset:
        dataset = Dataset(dataset, format)
        if format == DatasetFormat.NON_SPECIFIED:
            dataset.format = dataset_identifier.identify(dataset)
        return dataset

    def serialize(
            self,
            model_path: Union[str, Path, None] = None,
            dataset_path: Union[str, Path,
                                None] = None) -> Tuple[Model, Dataset]:
        if model_path is None:
            model_path = get_tmp_path()
        if dataset_path is None:
            dataset_path = get_tmp_path()
        model = model_serializer.serialize(self.model, Path(model_path))
        dataset = dataset_serializer.serialize(self.dataset, Path(dataset_path))

        # need to get the meta again after serialization
        return (self.__encapsulate_model(model.src, ModelFormat.NON_SPECIFIED,
                                         self.model.inputs, self.model.outputs,
                                         self.model.batch_dim),
                self.__encapsulate_dataset(dataset.src,
                                           DatasetFormat.NON_SPECIFIED))

    def evaluate(self, model, dataset) -> List[Metric]:
        return [e.evaluate(model, dataset) for e in self.evaluators]

    def add_model_transformer(self, model_transformer: ModelTransformer):
        self.model_transformers.append(model_transformer)

    def add_dataset_transformer(self, ds_transformer: DatasetTransformer):
        self.dataset_transformers.append(ds_transformer)

    def add_evaluator(self, evaluator: Evaluator):
        self.evaluators.append(evaluator)

    def dump(self) -> Dict:
        data = {
            "model":
                self.model.dump(),
            "dataset":
                self.dataset.dump(),
            "evaluators": [x.dump() for x in self.evaluators],
            "model_transformers": [x.dump() for x in self.model_transformers],
            "dataset_transformers": [
                x.dump() for x in self.dataset_transformers
            ]
        }
        return data

    def load(self, data: Dict):
        pass



from dataclasses import dataclass
from typing import Callable, Dict, Any, Type
import pickle

from ..dataset.dataset import Dataset
from ..model.model import Model


@dataclass
class Metric:
    name: str
    value: float


class Evaluator:
    def __init__(self, name: str, func: Callable, params: Dict[str, Any]):

        self.name = name
        self.func = func
        self.params = params

    def evaluate(self, model: Type[Model], dataset: Type[Dataset]) -> Metric:
        pass

    def dump(self) -> Dict[str, str]:
        data = {
            "func": pickle.dumps(self.func).decode("utf-8"),
            "params": pickle.dumps(self.params).decode("utf-8")
        }
        return {}



from typing import List
from importlib import import_module
from dataclasses import dataclass
from onnc.forest.core.runtime import AbstractRuntimeImplementation
from onnc.forest.core.options import Options, RuntimeSettings
from onnc.forest.core.runtime import make_runtime

from onnc.bench.core.model.model import Model
from onnc.bench.core.model import ModelFormat
import time

@dataclass
class ProxySettings(RuntimeSettings):
    pass


class ProxyOptions(Options):

    def __init__(self, loadable):
        self.settings: ProxySettings = ProxySettings()
        self.devices: List = []
        self.loadable: Model = loadable
        self._encapsulation: List = []

    @staticmethod
    def is_me(format) -> bool:
        return False # This runtime should be used manually


class RuntimeProxy(AbstractRuntimeImplementation):

    def __init__(self, options):
        self.options = options
        self._options_ = self.options._encapsulation[0]        
        self._runtime_ = make_runtime(self._options_)

    def launch(self, options: Options):
        self._runtime_.launch(self._options_)

    def load(self, options: Options):
        self._runtime_.load(self._options_)

    def bind_input(self, *args, **kwargs):
        self._runtime_.bind_input(*args, **kwargs)

    def bind_output(self, *args, **kwargs):
        self._runtime_.bind_output(*args, **kwargs)

    def bind_all_inputs(self):
        self._runtime_.bind_all_inputs()

    def bind_all_outputs(self):
        self._runtime_.bind_all_outputs()

    def materialize(self):
        self._runtime_.materialize()

    def write(self, data):
        self._runtime_.write(data)

    def run(self):
        self._runtime_.run()

    def read(self):
        self._runtime_.read()




from typing import List
from importlib import import_module
from dataclasses import dataclass
from onnc.forest.core.runtime import AbstractRuntimeImplementation

import zerorpc
import numpy as np
import time
import pickle

from onnc.forest.core.options import Options, RuntimeSettings
from onnc.forest.core.runtime import make_runtime

from onnc.bench.core.model.model import Model
from onnc.bench.core.model import ModelFormat
from onnc.forest.proxies import RuntimeProxy
from onnc.forest.proxies.zerorpc.utils import np_to_bytes, np_from_bytes


@dataclass
class ZeroRPCSettings(RuntimeSettings):
    protocol = 'tcp'
    host = '127.0.0.1'
    port = 20000


class ZeroRPCOptions(Options):

    def __init__(self, loadable):
        self.settings: ZeroRPCSettings = ZeroRPCSettings()
        self.devices: List = []
        self.loadable: Model = loadable
        self._encapsulation: List = []

    @staticmethod
    def is_me(format) -> bool:
        return False # This runtime should be used manually


class ZeroRPCProxy(RuntimeProxy):
    @staticmethod
    def is_me(options: Options) -> bool:
        return isinstance(options, ZeroRPCOptions)

    def __init__(self):
        pass

    def launch(self, options: Options):
        self._runtime_ = zerorpc.Client()
        self._runtime_.connect(f"{options.settings.protocol}://{options.settings.host}:{options.settings.port}")     
        self.options = options
        self._runtime_.init(pickle.dumps(options))
                
        options = pickle.dumps(options)
        self._runtime_.launch(options)

    def load(self, options: Options):
        options = pickle.dumps(options)
        self._runtime_.load(options)

    def bind_input(self, *args, **kwargs):
        self._runtime_.bind_input(*args, **kwargs)

    def bind_output(self, *args, **kwargs):
        self._runtime_.bind_output(*args, **kwargs)

    def bind_all_inputs(self):
        self._runtime_.bind_all_inputs()

    def bind_all_outputs(self):
        self._runtime_.bind_all_outputs()

    def materialize(self):
        self._runtime_.materialize()

    def write(self, data):
        self._runtime_.write(np_to_bytes(data))

    def run(self):
        self._runtime_.rpcrun()

    def read(self):
        return np_from_bytes(self._runtime_.read())




from __future__ import annotations

from typing import Any, Union, Dict, List, Callable, Tuple, Type
from collections.abc import Iterable
from pathlib import Path
from collections import OrderedDict
from loguru import logger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .layout import DataLayout

from ..common import get_class_name
from . import DatasetFormat


class Dataset:

    def __init__(self,
                 raw_dataset: Union[Iterable, Iterable[Path], object],
                 format: DatasetFormat = DatasetFormat.NON_SPECIFIED):
        """
        Do not put identify_format in Dataset. Make sure
        (MVC) control and model are separated .
        """
        if isinstance(raw_dataset, str):
            self.src = Path(raw_dataset)
        else:
            self.src =  raw_dataset
        self.y_label: OrderedDict[str, Union[int, float, str]] = OrderedDict()
        self.layout: Union[None, DataLayout] = None

        self.shape: Union[None, Tuple[int, ...]] = None

        self.format = format

    def clone_attributes(self, dataset: Type[Dataset]) -> None:
        """
        Clone user defined attributes from input

        Iter all user-defined attributes and check if they have values.
        If the attribute of the destination object has value, then skip
        the clone.
        Args:
            dataset Type[Dataset]: The dataset object to be cloned
        """

        _to_be_cloned: List = []
        for attr_name in dataset.__dict__:

            dst_value = getattr(self, attr_name)
            src_value = getattr(dataset, attr_name)

            if (isinstance(dst_value, OrderedDict)) or (isinstance(
                    dst_value, List)):
                if len(dst_value) == 0:
                    _to_be_cloned.append([attr_name, src_value])

            elif dst_value is None:
                _to_be_cloned.append([attr_name, src_value])

        for attr_name, src_value in _to_be_cloned:
            setattr(self, attr_name, src_value)

    def dump(self) -> Dict:
        """
        WARNING: the dumped dictionary can only be used for report
        since src can be path or class name (which is not a good design)
        """
        data = {
            "src": "",
            "y_label": self.y_label,
            "shape": self.shape,
            "layout": self.layout.dump() if self.layout is not None else None,
            "format": self.format.name
        }

        if isinstance(self.src, str):
            data["src"] = self.src
        elif isinstance(self.src, Path):
            data["src"] = str(self.src)
        else:
            data["src"] = get_class_name(self.src)

        return data

    def load(self, obj: Dict):
        raise NotImplementedError()


class ONNCDataset(Dataset):

    def __init__(self):
        super().__init__([])
        self.format = DatasetFormat.ONNC_DATASET


class NDARRAYDataset(Dataset):

    def __init__(self, raw_dataset: Iterable):
        super().__init__(raw_dataset)
        self.format = DatasetFormat.NDARRAY



from typing import List
from pathlib import Path
from abc import abstractmethod
import shutil
from dataclasses import dataclass
import zipfile

from loguru import logger

from .transformer import ModelTransformer
from .model import Model, Tensor
from . import ModelFormat, ModelDataType
from .identifier import identify
from ..common import get_tmp_path
from importlib import import_module
import numpy as np


class MetadataRetriverRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class ModelMeta:

    def __init__(self, inputs: List[Tensor], outputs: List[Tensor]):
        self.inputs = inputs
        self.outputs = outputs


class MetadataRetriver(metaclass=MetadataRetriverRegistry):

    FORMAT: ModelFormat = ModelFormat.NON_SPECIFIED

    @classmethod
    def is_me(cls, model: Model) -> bool:
        return identify(model) == cls.FORMAT

    @abstractmethod
    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        raise NotImplementedError("`retrieve_inputs` has to be implemented")

    @abstractmethod
    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        raise NotImplementedError("`retrieve_outputs` has to be implemented")

    def retrieve(self, model: Model) -> ModelMeta:
        inputs = self.retrieve_inputs(model)
        logger.debug(f"Inputs of {model.name}: {inputs}")
        outputs = self.retrieve_outputs(model)
        logger.debug(f"Outputs of {model.name}: {outputs}")
        return ModelMeta(inputs=inputs, outputs=outputs)


def retrieve_model_metadata(model: Model) -> ModelMeta:
    for metadataretriever in MetadataRetriver.REGISTRY:
        if metadataretriever.is_me(model):
            return metadataretriever().retrieve(model)

    raise NotImplementedError(f"Unalble to retrieve metadata of {model.format}")


class H5(MetadataRetriver):

    FORMAT = ModelFormat.H5

    from .dtype_map import TF_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        res = []
        for x in _model.inputs:
            shape = tuple(-1 if not s else s for s in x.shape)
            res.append(
                Tensor(name=x.name, shape=shape,
                       dtype=self.TF_map.map(x.dtype)))
        return res

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        res = []
        for x in _model.outputs:
            shape = tuple(-1 if not s else s for s in x.shape)
            res.append(
                Tensor(name=x.node.outbound_layer.name,
                       shape=shape,
                       dtype=self.TF_map.map(x.dtype)))

        return res


class ONNX(MetadataRetriver):

    FORMAT = ModelFormat.ONNX
    from .dtype_map import ONNX_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        import onnx
        onnx_model = onnx.load(model.src)

        input_all = [node.name for node in onnx_model.graph.input]
        input_initializer = [node.name for node in onnx_model.graph.initializer]
        net_feed_input = list(set(input_all) - set(input_initializer))

        res = []

        for input in onnx_model.graph.input:

            shape = tuple(
                xx.dim_value for xx in input.type.tensor_type.shape.dim)
            shape = tuple(-1 if int(s) == 0 else s for s in shape)

            if input.name in net_feed_input:
                res.append(
                    Tensor(input.name, shape,
                           self.ONNX_map.map(input.type.tensor_type.elem_type)))

        return res

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        import onnx
        model = onnx.load(model.src)
        res = []
        for model_output in model.graph.output:
            shape = tuple(
                s.dim_value for s in model_output.type.tensor_type.shape.dim)
            shape = tuple(-1 if int(s) == 0 else s for s in shape)
            res.append(
                Tensor(
                    model_output.name, shape,
                    self.ONNX_map.map(model_output.type.tensor_type.elem_type)))
        return res


class PTH(MetadataRetriver):

    FORMAT = ModelFormat.PTH
    from .dtype_map import NP_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        import torch
        _model = torch.load(model.src)
        tensor = list(_model.parameters())[  # type: ignore[union-attr]
            0].detach()
        t_name = "input" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tensor.shape, self.NP_map.map(tensor.dtype.type))
        ]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        import torch
        _model = torch.load(model.src)

        # type: ignore[union-attr]
        tensor = list(_model.parameters())[-1].detach()
        t_name = "output" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tensor.shape, self.NP_map.map(tensor.dtype.type))
        ]


class PB(MetadataRetriver):

    FORMAT = ModelFormat.PB
    from .dtype_map import TF_map

    # https://stackoverflow.com/questions/43517959/given-a-tensor-flow-model-graph-how-to-find-the-input-node-and-output-node-name
    def _load_graph(self, frozen_graph_filename):
        import tensorflow as tf
        with tf.io.gfile.GFile(frozen_graph_filename, "rb") as f:
            graph_def = tf.compat.v1.GraphDef()
            graph_def.ParseFromString(f.read())
        with tf.Graph().as_default() as graph:
            tf.import_graph_def(graph_def)

        ops = graph.get_operations()
        return ops

    def analyze_inputs_outputs(graph):

        inputs = []
        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)
            else:
                for input_tensor in op.inputs:
                    if input_tensor.op in outputs_set:
                        outputs_set.remove(input_tensor.op)
        outputs = list(outputs_set)
        return (inputs, outputs)

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        import tensorflow as tf
        ops = self._load_graph(model.src)
        inputs = []
        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)

        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.outputs[0].shape),
                   dtype=self.TF_map.map(x.outputs[0].dtype)) for x in inputs
        ]
        """
        >>> [x.size for x in graph_def.node[0].attr['shape'].shape.dim]
        [-1, 96, 96, 3]
        """

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        import tensorflow as tf
        ops = self._load_graph(model.src)
        inputs = []
        outputs_set = set(ops)

        for op in ops:
            if len(op.inputs) == 0 and op.type != 'Const':
                inputs.append(op)
            else:
                for input_tensor in op.inputs:
                    if input_tensor.op in outputs_set:
                        outputs_set.remove(input_tensor.op)

        outputs = list(outputs_set)

        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.outputs[0].shape),
                   dtype=self.TF_map.map(x.outputs[0].dtype)) for x in outputs
        ]


class SavedModel(MetadataRetriver):

    FORMAT = ModelFormat.SAVED_MODEL

    from .dtype_map import TF_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        assert type(model) == Model
        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)
        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype)) for x in _model.inputs
        ]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:

        from tensorflow import keras  # type: ignore[import]
        _model = keras.models.load_model(model.src)

        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1
                               for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype))
            for x in _model.inputs
            for x in _model.outputs
        ]


class ZippedSavedModel(SavedModel):

    FORMAT = ModelFormat.ZIPPED_SAVED_MODEL

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        temp = get_tmp_path()

        with zipfile.ZipFile(model.src,
                             'r') as zip_ref:  # type: ignore[arg-type]
            zip_ref.extractall(temp)
        return super().retrieve_inputs(Model(temp))

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        temp = get_tmp_path()
        with zipfile.ZipFile(model.src,
                             'r') as zip_ref:  # type: ignore[arg-type]
            zip_ref.extractall(temp)
        return super().retrieve_outputs(Model(temp))


class TFKerasModel(MetadataRetriver):

    FORMAT = ModelFormat.TF_KERAS_MODEL

    from .dtype_map import TF_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype)) for x in model.src.inputs
        ]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype)) for x in model.src.outputs
        ]


class KerasModel(MetadataRetriver):
    '''
    Keras 2.5.0 Serializer
    '''

    FORMAT = ModelFormat.KERAS_MODEL
    from .dtype_map import TF_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype)) for x in model.src.inputs
        ]  # type: ignore[union-attr]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        return [
            Tensor(name=x.name,
                   shape=tuple(xx if xx else -1 for xx in x.shape),
                   dtype=self.TF_map.map(x.dtype)) for x in model.src.outputs
        ]  # type: ignore[union-attr]


class PytorchModel(MetadataRetriver):
    """Use python MRO to check if it contains specific str"""

    FORMAT = ModelFormat.PT_NN_MODULE
    from .dtype_map import NP_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        _model = model.src
        tensor = list(_model.parameters())[  # type: ignore[union-attr]
            0].detach()
        t_name = "input" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tuple(x if x else -1 for x in tensor.shape),
                   self.NP_map.map(tensor.dtype.type))
        ]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        _model = model.src

        # type: ignore[union-attr]
        tensor = list(_model.parameters())[-1].detach()
        t_name = "output" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tuple(x if x else -1 for x in tensor.shape),
                   self.NP_map.map(tensor.dtype.type))
        ]


class TorchTracedModel(MetadataRetriver):

    FORMAT = ModelFormat.TORCH_TRACED
    from .dtype_map import NP_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:
        import torch
        _model = torch.jit.load(model.src)
        tensor = list(_model.parameters())[  # type: ignore[union-attr]
            0].detach()
        t_name = "input" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tuple(x if x else -1 for x in tensor.shape),
                   self.NP_map.map(tensor.dtype.type))
        ]

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        import torch
        _model = torch.jit.load(model.src)

        # type: ignore[union-attr]
        tensor = list(_model.parameters())[-1].detach()
        t_name = "output" if tensor.name == None else tensor.name
        tensor = tensor.numpy()
        return [
            Tensor(t_name, tuple(x if x else -1 for x in tensor.shape),
                   self.NP_map.map(tensor.dtype.type))
        ]


class TFLiteModel(MetadataRetriver):
    """Use python MRO to check if it contains specific str"""

    FORMAT = ModelFormat.TFLITE
    from .dtype_map import NP_map
    _dtype_map = NP_map

    def retrieve_inputs(self, model: Model) -> List[Tensor]:

        try:
            import tflite_runtime.interpreter as tflite
            interpreter = tflite.Interpreter(model_path=str(model.src))
        except ModuleNotFoundError:
            import tensorflow as tf
            interpreter = tf.lite.Interpreter(model_path=str(model.src))

        res = []
        for i in interpreter.get_input_details():
            res.append(
                Tensor(i["name"],
                       tuple(-1 if not s else int(s) for s in i["shape"]),
                       self._dtype_map.map(i['dtype'])))
        return res

    def retrieve_outputs(self, model: Model) -> List[Tensor]:
        try:
            import tflite_runtime.interpreter as tflite
            interpreter = tflite.Interpreter(model_path=str(model.src))
        except ModuleNotFoundError:
            import tensorflow as tf
            interpreter = tf.lite.Interpreter(model_path=str(model.src))

        res = []
        for i in interpreter.get_output_details():
            res.append(
                Tensor(i["name"],
                       tuple(-1 if not s else int(s) for s in i["shape"]),
                       self._dtype_map.map(i['dtype'])))
        return res



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict
import os

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dataset import Dataset



class DatasetOperation(ABC):
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    @abstractmethod
    def load(self):
        pass

    @abstractmethod
    def append(self, obj, label=None):
        pass

    @abstractmethod
    def remove(self):
        pass

    @abstractmethod
    def insert(self, index, obj):
        pass

    @abstractmethod
    def __getitem__(self, index: int):
        pass


class ONNCDatasetOperation(DatasetOperation):

    def load(self) -> None:
        pass

    def append_dir(self, path: Path, label_map: Dict = {},
                   suffix=['.jpg', '.jpeg', '.png', 'bmp']) -> None:
        for f in os.listdir(path):
            _f = path / Path(f)
            if _f.suffix.lower() in suffix:
                obj = _f.absolute()
                label = label_map[f] if f in label_map else None
                self.append(obj, label=label)

    def append(self, obj, label=None):
        self.dataset.src.append(obj)
        if label:
            self.dataset.y_label[str(len(self))] = label

    def remove(self, obj):
        pos = self.dataset.src.index(obj)
        self.dataset.src.remove(obj)
        del self.y_label[str(pos)]

    def insert(self, index, obj, label=None):
        self.dataset.src.insert(index, obj)
        self.dataset.y_label[id(obj)] = label

    def __getitem__(self, index: int):
        return self.dataset.src[index]

    def __len__(self):
        return len(self.dataset.src)

    def __iter__(self):
        return iter(self.dataset.src)


class ONNCImageDatasetOperation(ONNCDatasetOperation):
    def load(self):

        try:
            from PIL import Image
        except ImportError:
            print("Package Pillow/PIL is not found\nPlease install it using `pip install Pillow`")
        import numpy as np

        data = []
        for f in self.dataset.src:  # list of path
            image = Image.open(str(f))
            data.append(np.asarray(image))
        self.dataset.src = data



from typing import Union, List
from abc import abstractmethod
from pathlib import Path
import shutil
import numpy as np
from loguru import logger

from .identifier import DatasetFormat, identify
from ..common import get_tmp_path
from .transformer import DatasetTransformer
from .dataset import Dataset


class SerializerRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):

        new_cls = type.__new__(cls, name, bases, attrs)

        cls.REGISTRY.append(new_cls)
        return new_cls


class Serializer(DatasetTransformer, metaclass=SerializerRegistry):

    FORMAT: Union[None, DatasetFormat] = None

    @classmethod
    def is_me(cls, dataset: Dataset) -> bool:
        return identify(dataset) == cls.FORMAT

    @abstractmethod
    def transform(self, dataset: Dataset) -> Dataset:
        pass

    def serialize(self, dataset: Dataset, dest: Path) -> Dataset:
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        return self.transform(dataset)


def serialize(dataset: Dataset, dest: Path) -> Dataset:
    for serializer in Serializer.REGISTRY:
        if serializer.is_me(dataset):
            return serializer().serialize(dataset, dest)

    raise NotImplementedError(f"Unalble to identify {dataset.src}")


class FileSerializer(Serializer):

    FORMAT: DatasetFormat = DatasetFormat.NON_SPECIFIED

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')

        if dataset.src.is_file():
            dest = dest.with_suffix(dataset.src.suffix)
            shutil.copy(dataset.src, dest)
        else:
            raise Exception(
                f'src [`{dataset.src}`] should be a path to a file.')

        return Dataset(dest)

    def serialize(self, dataset: Dataset, dest: Path) -> Dataset:
        # return super().serialize(model, dest)
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        m = self.transform(dataset)
        return m


class DirSerializer(Serializer):

    FORMAT: DatasetFormat = DatasetFormat.NON_SPECIFIED

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')

        if Path(dest).exists() and Path(dest).is_dir():
            raise Exception(f'dest `{dest}` should be a file path.')
        elif not Path(dest).exists:
            if dest[-1] in ['\\', '/']:
                raise Exception(f"dest `{dest}` should file path, not a dir")

        if dataset.src.is_dir():
            # Zip the dir if the model is in dir form
            dest = dest.with_suffix('.zip')
            tmp_path = Path(get_tmp_path()[:-1])
            shutil.make_archive(str(tmp_path), 'zip', dataset.src)
            shutil.move(str(tmp_path) + '.zip', str(dest))
        else:
            raise Exception(
                f'dataset source `{dataset.src}` should be a path to a dir.')

        return Dataset(dest)

    def serialize(self, dataset: Dataset, dest: Path) -> Dataset:
        # return super().serialize(model, dest)
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        m = self.transform(dataset)
        return m


class NONE(Serializer):

    FORMAT = DatasetFormat.NONE

    def transform(self, dataset: Dataset):
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        np.save(dest, np.zeros([1, 3, 3, 3]))
        shutil.move(str(dest.with_suffix('.npy')), str(dest))
        return Dataset(dest)


class PB(FileSerializer):

    FORMAT = DatasetFormat.PB


class NPY(FileSerializer):

    FORMAT = DatasetFormat.NPY


class NPYDIR(DirSerializer):

    FORMAT = DatasetFormat.NPYDIR


class NPZ(FileSerializer):

    FORMAT = DatasetFormat.NPZ


class NDARRAY(Serializer):

    FORMAT = DatasetFormat.NDARRAY

    def transform(self, dataset: Dataset):
        assert isinstance(dataset.src, np.ndarray)
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        np.save(dest, dataset.src)
        shutil.move(str(dest.with_suffix('.npy')), str(dest))
        return Dataset(dest)


class TORCH_DATASET(Serializer):

    FORMAT = DatasetFormat.TORCH_DATASET

    def transform(self, dataset: Dataset):
        import numpy
        from torch import Tensor
        from torch.utils.data import Dataset as torchDataset
        assert isinstance(dataset.src, torchDataset)
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        if isinstance(dataset.src.data, numpy.ndarray):
            dataset_npy = Dataset(dataset.src.data)
        elif isinstance(dataset.src.data, Tensor):
            dataset_npy = Dataset(
                dataset.src.data.numpy())  # type: ignore[union-attr]

        # Handle NHWC data (Guess, might fail)
        if len(dataset_npy.src.shape) == 4 and dataset_npy.src.shape[-1] <= 3:
            dataset_npy.src = dataset_npy.src.transpose((0, 3, 1, 2))

        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)


class TORCH_DATALOADER(Serializer):

    FORMAT = DatasetFormat.TORCH_DATALOADER

    def transform(self, dataset: Dataset):
        from torch.utils.data import DataLoader
        assert isinstance(dataset.src, DataLoader)
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        dataset_torch = Dataset(dataset.src.dataset)  # type: ignore[union-attr]
        torchdataset_serializer = TORCH_DATASET()
        return torchdataset_serializer.serialize(dataset_torch, dest)


class KERAS_DATASET(Serializer):

    FORMAT = DatasetFormat.KERAS_DATASET

    def transform(self, dataset: Dataset):
        assert isinstance(dataset.src, tuple)
        dest = self.get_param('dest')  # type: ignore[attr-defined]
        # Keras dataset is a 2x2 metric of tuple
        # ds[0] is for training , ds[0][0] is x and ds[0][1] is y
        # ds[1] is for testing  , ds[1][0] is x and ds[1][1] is y
        # each record is a np.ndarray object
        # hence, we use x of testing data here
        dataset_npy = Dataset(dataset.src[1][0])  #type: ignore[index]
        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)


class TFDS_PREFETCH(Serializer):

    FORMAT = DatasetFormat.TFDS_PREFETCH
    SAMPLE_NUM = 1000
    SHUFFLE_SIZE = 1024

    def transform(self, dataset: Dataset):
        import tensorflow_datasets as tfds  # type: ignore[import]
        import tensorflow as tf
        assert isinstance(dataset.src, tf.data.Dataset)
        LABEL_KEYS = ['label']

        dest = self.get_param('dest')  # type: ignore[attr-defined]
        try:
            sample_num = self.get_param(
                'sample_num')  #type: ignore[attr-defined]

            if not sample_num:
                logger.warning('Parameter `sample_num` is not set, '
                               'use default value `1000`')
                sample_num = self.SAMPLE_NUM
        except Exception as e:
            logger.warning(
                'Parameter `sample_num` is not set, use default value `1000`')
            sample_num = self.SAMPLE_NUM

        ds = [
            x for x in dataset.src.shuffle(  # type: ignore[union-attr]
                self.SHUFFLE_SIZE).take(sample_num).as_numpy_iterator()
        ]

        keys = list(ds[0].keys())
        if len(keys) > 1:
            try:
                for key in LABEL_KEYS:
                    keys.remove(key)
            except ValueError:
                pass

        if len(keys) > 1:
            try:
                data_key = self.get_param(
                    'data_key')  #type: ignore[attr-defined]
                if not data_key:
                    logger.warning(f'Parameter `data_key` is not set, '
                                   'use default value `{keys[0]}`')
                    data_key = keys[0]
            except Exception as e:
                logger.warning(f'Parameter `data_key` is not set, '
                               'use default value `{keys[0]}`')
                data_key = keys[0]
        else:
            data_key = keys[0]

        dataset_npy = Dataset(np.stack([x[data_key] for x in ds]))

        ndarray_serializer = NDARRAY()
        return ndarray_serializer.serialize(dataset_npy, dest)



from typing import List
from importlib import import_module
from dataclasses import dataclass
from onnc.forest.core.runtime import AbstractRuntimeImplementation

from onnc.forest.core.options import Options, RuntimeSettings
from onnc.bench.core.model.model import Model
from onnc.bench.core.model import ModelFormat


@dataclass
class ONNXSettings(RuntimeSettings):
    ep_list = ['CPUExecutionProvider']
    intra_op_num_threads = 0


class ONNXOptions(Options):
    model_format = ModelFormat.ONNX

    def __init__(self, loadable):
        self.settings: ONNXSettings = ONNXSettings()
        self.devices: List = ["CPU"]
        self.loadable: Model = loadable


class ONNXRuntime(AbstractRuntimeImplementation):

    @staticmethod
    def is_me(options: Options) -> bool:
        return isinstance(options, ONNXOptions)


    def __init__(self):
        self.__ort = import_module('onnxruntime')

    def launch(self, options: ONNXOptions):
        pass

    def load(self, options: ONNXOptions):
        sess_options = self.__ort.SessionOptions()
        sess_options.intra_op_num_threads = (
            options.settings.intra_op_num_threads)
        model_path = str(options.loadable.src)
        self._sess = self.__ort.InferenceSession(
            model_path,
            sess_options=sess_options,
            providers=options.settings.ep_list)
        self._inputs_name = [input.name for input in self._sess.get_inputs()]
        self._outputs_name = [
            output.name for output in self._sess.get_outputs()
        ]

    def bind_input(self, name, obj):
        pass

    def bind_output(self, name, obj):
        pass

    def bind_all_inputs(self):
        pass

    def bind_all_outputs(self):
        pass

    def materialize(self):
        pass

    def write(self, data):
        self._input_data = data

    def run(self):
        self.out_buf = self._sess.run(self._outputs_name, {
            name: data
            for name, data in zip(self._inputs_name, self._input_data)
        })

    def read(self):
        return self.out_buf



from typing import List
from pathlib import Path
from abc import abstractmethod
import shutil
from loguru import logger

from .transformer import ModelTransformer
from .model import Model
from . import ModelFormat
from .identifier import identify
from ..common import get_tmp_path


class SerializerRegistry(type):

    REGISTRY: List = []

    def __new__(cls, name, bases, attrs):

        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Serializer(ModelTransformer, metaclass=SerializerRegistry):

    FORMAT = None

    @classmethod
    def is_me(cls, model: Model) -> bool:
        return identify(model) == cls.FORMAT

    @abstractmethod
    def transform(self, model: Model) -> Model:
        raise NotImplementedError("`transform` has to be implemented")

    def serialize(self, model: Model, dest: Path) -> Model:
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        m = self.transform(model)
        return m


def serializer_selector(model: Model) -> Serializer:
    for serializer in Serializer.REGISTRY:
        if serializer.is_me(model):
            return serializer
    raise NotImplementedError(f"Unalble to serialize {model}")


def serialize(model: Model, dest: Path) -> Model:
    for serializer in Serializer.REGISTRY:
        if serializer.is_me(model):
            return serializer().serialize(model, dest)

    raise NotImplementedError(f"Unalble to serialize {model}")


class FileSerializer(Serializer):

    FORMAT: ModelFormat = ModelFormat.NON_SPECIFIED

    def transform(self, model: Model):
        dest = self.get_param('dest')

        if model.src.is_file():
            dest = dest.with_suffix(model.src.suffix)
            shutil.copy(model.src, dest)
        else:
            raise Exception(f'src [`{model.src}`] should be a path to a file.')

        return Model(dest)

    def serialize(self, model: Model, dest: Path) -> Model:
        # return super().serialize(model, dest)
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        m = self.transform(model)
        return m


class DirSerializer(Serializer):

    FORMAT: ModelFormat = ModelFormat.NON_SPECIFIED

    def transform(self, model: Model):
        dest = self.get_param('dest')

        if Path(dest).exists() and Path(dest).is_dir():
            raise Exception(f'dest `{dest}` should be a file path.')
        elif not Path(dest).exists:
            if dest[-1] in ['\\', '/']:
                raise Exception(f"dest `{dest}` should file path, not a dir")

        if model.src.is_dir():
            # Zip the dir if the model is in dir form
            dest = dest.with_suffix('.zip')
            tmp_path = Path(get_tmp_path()[:-1])
            shutil.make_archive(str(tmp_path), 'zip', model.src)
            shutil.move(str(tmp_path) + '.zip', str(dest))
        else:
            raise Exception(
                f'model source `{model.src}` should be a path to a dir.')

        return Model(dest)

    def serialize(self, model: Model, dest: Path) -> Model:
        # return super().serialize(model, dest)
        self.add_param('dest', dest)  # type: ignore[attr-defined]
        m = self.transform(model)
        return m


class H5(FileSerializer):

    FORMAT = ModelFormat.H5


class ONNX(FileSerializer):

    FORMAT = ModelFormat.ONNX


class PTH(FileSerializer):

    FORMAT = ModelFormat.PTH


class PB(FileSerializer):

    FORMAT = ModelFormat.PB


class TorchTraced(FileSerializer):

    FORMAT = ModelFormat.TORCH_TRACED


class TFLITE(FileSerializer):

    FORMAT = ModelFormat.TFLITE


class SavedModel(DirSerializer):

    FORMAT = ModelFormat.SAVED_MODEL


class ZippedSavedModel(FileSerializer):

    FORMAT = ModelFormat.ZIPPED_SAVED_MODEL


class TFKerasModel(Serializer):
    '''
    '''

    FORMAT = ModelFormat.TF_KERAS_MODEL

    def transform(self, model: Model):
        dest = self.get_param('dest')

        # TF model.save uses file ext to determine the format to
        # be saved. if not specified, SavedModel willl be used.
        model.src.save(dest.with_suffix('.h5'))
        shutil.move(dest.with_suffix('.h5'), str(dest))
        return Model(dest)


class KerasModel(Serializer):
    '''
    Keras 2.5.0 Serializer
    '''

    FORMAT = ModelFormat.KERAS_MODEL

    def transform(self, model: Model):
        dest = self.get_param('dest')
        model.src.save(dest.with_suffix('.h5'))
        shutil.move(dest.with_suffix('.h5'), str(dest))
        return Model(dest)


class TF_Session(Serializer):
    '''
    Keras 2.5.0 Serializer
    '''

    FORMAT = ModelFormat.TF_SESSION

    def transform(self, model: Model):
        import tensorflow

        dest = self.get_param('dest')

        saver = tensorflow.train.Saver()

        sess = model.src
        sess.run(tensorflow.initialize_all_variables())

        model.src.save(dest.with_suffix('.h5'))
        shutil.move(dest.with_suffix('.h5'), str(dest))
        return Model(dest)


class PytorchModel(Serializer):
    """Use python MRO to check if it contains specific str"""

    FORMAT = ModelFormat.PT_NN_MODULE

    def transform(self, model: Model):
        import torch

        dest = self.get_param('dest')

        if not model.inputs:
            raise Exception(
                "PytorchModel requires input shpae, please add an input tensor in the Model"
            )

        for input_tensor in model.inputs:
            try:
                if not all(isinstance(x, int) for x in input_tensor.shape):
                    raise Exception(f"Parameter `shape` must be List[int]")
            except Exception as e:
                raise Exception("Parameter `shape` must be List[int]")

        dummy_inputs = []
        if not len(model.inputs) > 0:
            raise Exception(
                "Pytorch model has to have at least one input tensor")

        for input_tensor in model.inputs:
            shape = list(input_tensor.shape)
            shape[0] = 1
            shape = tuple(shape)
            dummy_inputs.append(torch.rand(*shape))

        # torch.onnx.export(model.src, tuple(dummy_inputs), dest,
        #                   input_names=[x.name for x in model.inputs],
        #                   output_names=[x.name for x in model.outputs])

        try:
            traced_module = torch.jit.trace(model.src, tuple(dummy_inputs))
        except ValueError as ve:
            logger.error("Unable to save the Pytorch model. "
                         "Specify `inputs` in project.add_model() "
                         "may solve this issue.")
            raise ve

        torch.jit.save(traced_module, dest)

        return Model(dest)



from functools import singledispatch
from onnc.forest.runtimes.tensorrt.runtime import TensorRTRuntime
from onnc.forest.runtimes.openvino.runtime import OpenvinoRuntime


def auto_bind(runtime):
    for rt in runtime.loaded_runtimes:
        try:
            _auto_bind(rt)
        except:
            pass


@singledispatch
def _auto_bind(runtime):
    raise NotImplementedError(f"Runtime {runtime} not support auto_bind")


@_auto_bind.register
def _(runtime: TensorRTRuntime):
    from onnc.forest.runtimes.tensorrt.utils import allocate_buffers
    runtime._inputs, runtime._outputs, runtime._bindings, _ = \
        allocate_buffers(runtime._cuda_engine)

@_auto_bind.register
def _(runtime: OpenvinoRuntime):
    runtime.bind_input(None,None)


from typing import Any, Dict
from abc import ABC, abstractmethod
import pickle

from .model import Model


class ModelTransformer():

    NAME = 'ModelTransformer'

    def __init__(self):
        self.params = {}

    def transform(self, model: Model) -> object:
        raise NotImplementedError("`transform` has to be implemented")

    def add_param(self, key: str, value: Any) -> None:
        self.params[key] = value

    def get_param(self, key: str) -> Any:
        return self.params[key]

    def dump(self) -> str:
        return pickle.dumps(self).decode("utf-8")



from pathlib import Path
from typing import Dict, List, Any, Callable, Union, Set
import os
import json
import shutil

from loguru import logger

from onnc.bench.core.deployment import Deployment
from ..dataset.dataset import Dataset
from ..model.model import Model
from ..modelpackage import ModelPackage
from onnc.bench.core.common import get_tmp_path
from . import Compilation
from .nnuxe import NNUXEBuilder


class NNUXEDockerBuilder(NNUXEBuilder):
    BUILDER_NAME = "NNUXEDockerBuilder"

    def __init__(self,
                 image_name='nnuxe/nnuxe/nnuxe_image',
                 image_tag="latest"):
        super().__init__()
        self.image_name = image_name
        self.image_tag = image_tag

    def _compile(self, model_name, model_path: str, sample_path: str,
                 params_path: str, output_path: str,
                 local_nnuxe: bool = False):
        build_path = '/build'
        build_tmp = get_tmp_path()
        os.makedirs(build_tmp, exist_ok=True)

        shutil.copy(model_path, os.path.join(build_tmp, 'model.file'))
        shutil.copy(sample_path, os.path.join(build_tmp, 'sample.file'))
        shutil.copy(params_path, os.path.join(build_tmp, 'params.file'))

        os.makedirs(output_path, exist_ok=True)
        output_path = os.path.join(output_path, model_name)

        cmd = [f'docker run -it -v {build_tmp}:{build_path}',
               f'-e PYTHONPATH="/root/nnuxe/nnuxe/external/tvm/python::/root/nnuxe/"',
               f'-e NNUXE_PATH="/root/nnuxe/"',
               f'-e TVM_HOME="/root/nnuxe/nnuxe/external/tvm"',
               f'{self.image_name}:{self.image_tag}',
               f'python /root/nnuxe/cli/compile.py --model {build_path}/model.file',
               f'--sample {build_path}/sample.file',
               f'--params {build_path}/params.file --output {build_path}/out']
        cmd = ' '.join(cmd)

        logger.debug(cmd)

        os.system(cmd)

        with open(os.path.join(build_tmp, 'out', 'report.json'), 'r') as f:
            report = json.load(f)

        shutil.move(os.path.join(build_tmp, 'out'), output_path)

        return report



from enum import Enum

# from .layout import DataLayout
# from .dataset import Dataset
# from .identifier import Identifier
# from .transformer import DatasetTransformer


__DATASET__ = '__DATASET__'


class DatasetFormat(Enum):
    NON_SPECIFIED = 0
    UNKNOWN = 1
    NONE = 2

    NPY = 11
    NPYDIR = 12
    NPZ = 13
    NDARRAY = 14
    NP_MEMMAP = 15
    NPZ_OBJECT = 16

    PB = 17

    ONNC_DATASET = 51

    TORCH_DATASET = 100
    TORCH_DATALOADER = 101

    KERAS_DATASET = 200
    TFDS_PREFETCH = 201 # tensorflow.python.data.ops.dataset_ops.PrefetchDataset




from __future__ import annotations
from typing import Any, Dict, List, Callable, Type
import types
import pickle

from .dataset import Dataset, NDARRAYDataset

from . import __DATASET__


class DatasetTransformer():

    NAME = 'DatasetTransformer'

    def __init__(self):
        self.params = {}

    def transform(self, dataset: Type[Dataset]) -> object:
        return dataset

    def add_param(self, key: str, value: Any) -> None:
        self.params[key] = value

    def get_param(self, key: str) -> Any:
        return self.params[key]

    def set_transform(self, func: Callable, args: List = [],
                      kwargs: Dict = {}) -> None:

        """Set the transform func as given callback

        Args:
            func (Callable): preprocessor function, the output of the this
                             function has to be the preprocessed dataset.
            args (List): args for the preprocessor function, use `__DATASET__`
                         to indicate the dataset argument position

            kwargs (Dict): keword args for the preprocessor function

        Example:

            def _preprocess(dataset, min_val, max_val, enhancement=False):
              ....
              ....

            set_preprocessor(_preprocess,
                            ['__DATASET__', 1, 100],
                            {'enhancement': True}
                            )
        """

        def _transform(_self, dataset):

            # adapt __DATASET__
            for idx, i in enumerate(args):
                if i == __DATASET__:
                    args[idx] = dataset.src

            dataset.src = func(*args, **kwargs)
            return dataset

        setattr(self, 'transform', types.MethodType(_transform, self))

    def dump(self) -> bytes:
        return pickle.dumps(self)


class DatasetFormatTransformer(DatasetTransformer):
    NAME = 'DatasetFormatTransformer'


class ONNCDatasetToNumpy(DatasetFormatTransformer):
    NAME = 'ONNCDatasetToNumpy'

    def transform(self, dataset: Type[Dataset]) -> Type[Dataset]:
        import numpy as np

        # check if all elements are ndarray
        for obj in dataset.src:
            assert isinstance(obj, np.ndarray)

        data = np.concatenate(dataset.src, axis=0)

        dst_dataset = NDARRAYDataset(data)
        dst_dataset.clone_attributes(dataset)
        dst_dataset.src = data

        return dst_dataset




import logging
import threading
from collections import deque
from typing import Dict, Set
import xmltodict

def retrieve_input_output_names(openvino_ir_xml_path):
    with open(openvino_ir_xml_path, 'r') as myfile:
        obj = xmltodict.parse(myfile.read())
        input = []
        output = []

        for dic in obj["net"]["layers"]["layer"]:
            if dic.get("@type") == "Parameter":
                input.append(dic["@name"])
            elif dic.get("@type") == "Result":
                output.append(dic["@name"].split("/")[0])
        return input, output



class AsyncPipeline:

    def __init__(self,
                 ie,
                 xml_model_path,
                 plugin_config,
                 device='CPU',
                 max_num_requests=1):

        self.net = ie.read_network(xml_model_path)
        self.logger = logging.getLogger()

        self.logger.info('Loading network to {} plugin...'.format(device))
        self.exec_net = ie.load_network(network=self.net,
                                        device_name=device,
                                        config=plugin_config,
                                        num_requests=max_num_requests)
        if max_num_requests == 0:
            # ExecutableNetwork doesn't allow creation of additional InferRequests. Reload ExecutableNetwork
            # +1 to use it as a buffer of the pipeline
            self.exec_net = ie.load_network(
                network=self.net,
                device_name=device,
                config=plugin_config,
                num_requests=len(self.exec_net.requests) + 1)
        self.empty_requests = deque(self.exec_net.requests)
        self.completed_request_results = {}
        self.callback_exceptions = {}
        self.event = threading.Event()

    def inference_completion_callback(self, status, callback_args):
        try:
            request, id, meta, preprocessing_meta = callback_args
            if status != 0:
                raise RuntimeError(
                    'Infer Request has returned status code {}'.format(status))
            raw_outputs = {
                key: blob.buffer for key, blob in request.output_blobs.items()
            }
            self.completed_request_results[id] = (raw_outputs, meta,
                                                  preprocessing_meta)
            self.empty_requests.append(request)
        except Exception as e:
            self.callback_exceptions.append(e)
        self.event.set()

    def submit_data(self, inputs, id, meta):
        request = self.empty_requests.popleft()
        if len(self.empty_requests) == 0:
            self.event.clear()
        request.set_completion_callback(
            py_callback=self.inference_completion_callback,
            py_data=(request, id, meta, None))
        request.async_infer(inputs=inputs)

    def get_raw_result(self, id):
        if id in self.completed_request_results:
            return self.completed_request_results.pop(id)
        return None

    def get_result(self, id):
        result = self.get_raw_result(id)
        if result:
            raw_result, meta, preprocess_meta = result
            # return self.model.postprocess(raw_result, preprocess_meta), meta
            return raw_result, meta
        return None

    def is_ready(self):
        return len(self.empty_requests) != 0

    def has_completed_request(self):
        return len(self.completed_request_results) != 0

    def await_all(self):
        for request in self.exec_net.requests:
            request.wait()

    def await_any(self):
        if len(self.empty_requests) == 0:
            self.event.wait()


def parse_devices(device_string):
    colon_position = device_string.find(':')
    if colon_position != -1:
        device_type = device_string[:colon_position]
        if device_type == 'HETERO' or device_type == 'MULTI':
            comma_separated_devices = device_string[colon_position + 1:]
            devices = comma_separated_devices.split(',')
            for device in devices:
                parenthesis_position = device.find(':')
                if parenthesis_position != -1:
                    device = device[:parenthesis_position]
            return devices
    return (device_string,)


def parse_value_per_device(devices: Set[str],
                           values_string: str) -> Dict[str, int]:
    """Format: <device1>:<value1>,<device2>:<value2> or just <value>"""
    values_string_upper = values_string.upper()
    result = {}
    device_value_strings = values_string_upper.split(',')
    for device_value_string in device_value_strings:
        device_value_list = device_value_string.split(':')
        if len(device_value_list) == 2:
            if device_value_list[0] in devices:
                result[device_value_list[0]] = int(device_value_list[1])
        elif len(device_value_list) == 1 and device_value_list[0] != '':
            for device in devices:
                result[device] = int(device_value_list[0])
        elif device_value_list[0] != '':
            raise RuntimeError(f'Unknown string format: {values_string}')
    return result

def get_user_config(flags_d: str, flags_nstreams: str,
                    flags_nthreads: int) -> Dict[str, str]:
    config = {}

    devices = set(parse_devices(flags_d))

    device_nstreams = parse_value_per_device(devices, flags_nstreams)
    for device in devices:
        if device == 'CPU':  # CPU supports a few special performance-oriented keys
            # limit threading for CPU portion of inference
            if flags_nthreads:
                config['CPU_THREADS_NUM'] = str(flags_nthreads)

            config['CPU_BIND_THREAD'] = 'NO'

            # for CPU execution, more throughput-oriented execution via streams
            config['CPU_THROUGHPUT_STREAMS'] = str(device_nstreams[device]) \
                if device in device_nstreams else 'CPU_THROUGHPUT_AUTO'
        elif device == 'GPU':
            config['GPU_THROUGHPUT_STREAMS'] = str(device_nstreams[device]) \
                if device in device_nstreams else 'GPU_THROUGHPUT_AUTO'
            if 'MULTI' in flags_d and 'CPU' in devices:
                # multi-device execution with the CPU + GPU performs best with GPU throttling hint,
                # which releases another CPU thread (that is otherwise used by the GPU driver for active polling)
                config['GPU_PLUGIN_THROTTLE'] = '1'
    return config


from pathlib import Path
from typing import Any, List, Union, Dict
from onnc.bench.core.model.model import Model
from onnc.bench.core.dataset.dataset import Dataset
from onnc.bench.core.dataset.layout import DataLayout


class Compilation:

    @classmethod
    def get_model_meta(cls, model: Model, inputs_as_nchw):
        return {
            "batch_dim": model.batch_dim,
            "inputs": [x.dump() for x in model.inputs],
            "inputs_as_nchw": inputs_as_nchw,
            "outputs": [x.dump() for x in model.outputs],
            "format": model.format.name
        }

    @classmethod
    def get_dataset_meta(cls, dataset: Dataset):

        dataset_meta = {"shape": dataset.shape, "format": dataset.format.name}

        if isinstance(dataset.layout, DataLayout):
            dataset_meta["layout"] = dataset.layout.dump()

        return dataset_meta

    def __init__(self,
                 model: Model,
                 dataset: Dataset,
                 inputs_as_nchw: Union[str, bool, None, List[str]] = None):
        if not (isinstance(model.src, str) or isinstance(model.src, Path)):
            raise TypeError("model.src should be str or path, "
                            f"not {type(model.src)}")
        if not (isinstance(dataset.src, str) or isinstance(dataset.src, Path)):
            raise TypeError("dataset.src should be str or path, "
                            f"not {type(dataset.src)}")
        self.compilation_id: str  # SaaS compilation_id
        self.model_path = model.src
        self.sample_path = dataset.src
        self.model_meta = self.get_model_meta(model, inputs_as_nchw)
        self.sample_meta = self.get_dataset_meta(dataset)
        self.calibrator_params: Dict[str, Any]
        self.compiler_params: Dict[str, Any]


from .onnc_saas import ONNCSaaSBuilder
from .nnuxe import NNUXEBuilder
from .nnuxe_docker import NNUXEDockerBuilder



"""Setup script for realpython-reader"""

import os.path
from setuptools import setup, find_packages

# The directory containing this file
HERE = os.path.abspath(os.path.dirname(__file__))

# The text of the README file
with open(os.path.join(HERE, "README.md")) as fid:
    README = fid.read()

# This call to setup() does all the work
setup(name="onnc-bench",
      version="4.1.4",
      description="ONNC-bench is a Python wrapper of ONNC",
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://www.skymizer.com",
      author="The Skymizer Team",
      author_email="hello@skymizer.com",
      license="Apache License 2.0",
      packages=find_packages(),
      package_data={"onnc": ["*"]},
      data_files=[],
      install_requires=["requests", "numpy", "onnx", "loguru", "sentry-sdk"])



import time
from typing import List
from importlib import import_module
from dataclasses import dataclass
import pickle

import zerorpc

from onnc.forest.core.options import Options, RuntimeSettings
from onnc.forest.core.runtime import make_runtime

from onnc.bench.core.model.model import Model
from onnc.bench.core.model import ModelFormat
from onnc.forest.core.runtime import make_runtime
from onnc.forest.proxies import RuntimeProxy
from onnc.forest.proxies.zerorpc.utils import np_to_bytes, np_from_bytes


@dataclass
class ZeroRPCServerSettings(RuntimeSettings):
    protocol = 'tcp'
    host = '0.0.0.0'
    port = 20000


class ZeroRPCServerOptions(Options):
    def __init__(self, loadable):
        self.settings: ZeroRPCServerSettings = ZeroRPCServerSettings()
        self.devices: List = []
        self.loadable: Model = loadable
        self._encapsulation: List = []

    @staticmethod
    def is_me(format) -> bool:
        return False # This runtime should be used manually


class ZeroRPCServerProxy(RuntimeProxy):
    def __init__(self, options):
        pass

    def init(self, options):
        options = pickle.loads(options)
        self.options = options
        self._options_ = self.options._encapsulation[0]        
        self._runtime_ = make_runtime(self._options_)

    def launch(self, options: Options):
        self._runtime_.launch(self._options_)

    def load(self, options: Options):
        self._runtime_.load(self._options_)

    def bind_input(self, *args, **kwargs):
        self._runtime_.bind_input(*args, **kwargs)

    def bind_output(self, *args, **kwargs):
        self._runtime_.bind_output(*args, **kwargs)

    def bind_all_inputs(self):
        self._runtime_.bind_all_inputs()

    def bind_all_outputs(self):
        self._runtime_.bind_all_outputs()

    def materialize(self):
        self._runtime_.materialize()

    def write(self, data):
        data = np_from_bytes(data)
        self._runtime_.write(data)

    def rpcrun(self):
        self._runtime_.run()

    def run(self):
        raise Exception('ZeroRPC cannot use `run` as method name')

    def read(self):
        return np_to_bytes(self._runtime_.read())

if __name__ == "__main__":
    opt = ZeroRPCServerOptions(None)
    s = zerorpc.Server(ZeroRPCServerProxy(opt))
    s.bind(f"{opt.settings.protocol}://{opt.settings.host}:{opt.settings.port}")
    s.run()


from dataclasses import dataclass
from onnc.bench.core.model import ModelFormat, ModelDataType
from onnc.bench.utils import Singleton


@dataclass
class Base_map(metaclass=Singleton):
    _map = {}
    _inv_map = {}
    _inv_name_map = {}

    @classmethod
    def map(cls, key):
        return cls._map[key]

    @classmethod
    def inv_map(cls, key):
        return cls._inv_map[key]

    @classmethod
    def inv_name_map(cls, key):
        return cls._inv_name_map[key]

    @classmethod
    def str_map(cls, key):
        return cls._str_map[key]



@dataclass
class TORCH_map(Base_map):

    import torch
    _map = {
        torch.float: ModelDataType.FP32,
        torch.float16: ModelDataType.FP16,
        torch.float32: ModelDataType.FP32,
        torch.float64: ModelDataType.FP64,
        torch.uint8: ModelDataType.UINT8,
        torch.int8: ModelDataType.INT8,
        torch.int16: ModelDataType.INT16,
        torch.int32: ModelDataType.INT32,
        torch.int64: ModelDataType.INT64,
        torch.bool: ModelDataType.BOOL,
        torch.complex128: ModelDataType.COMPLEX128,
        torch.bfloat16: ModelDataType.BFLOAT16
    }
    _str_map = {str(k):v.name for k, v in _map.items()}
    _inv_map = {v: k for k, v in _map.items()}
    _inv_name_map = {v.name: k for k, v in _map.items()}


@dataclass
class TF_map(Base_map):

    import tensorflow as tf
    _map = {
        tf.float32: ModelDataType.FP32,
        tf.uint8: ModelDataType.UINT8,
        tf.int8: ModelDataType.INT8,
        tf.uint16: ModelDataType.UINT16,
        tf.int16: ModelDataType.INT16,
        tf.int32: ModelDataType.INT32,
        tf.int64: ModelDataType.INT64,
        tf.string: ModelDataType.STR,
        tf.bool: ModelDataType.BOOL,
        tf.float16: ModelDataType.FP16,
        tf.float64: ModelDataType.FP64,
        tf.uint32: ModelDataType.UINT32,
        tf.uint64: ModelDataType.UINT64,
        tf.complex128: ModelDataType.COMPLEX128,
        tf.bfloat16: ModelDataType.BFLOAT16
    }
    _str_map = {str(k):v.name for k, v in _map.items()}
    _inv_map = {v: k for k, v in _map.items()}
    _inv_name_map = {v.name: k for k, v in _map.items()}


@dataclass
class ONNX_map(Base_map):
    _map = {
        1: ModelDataType.FP32,
        2: ModelDataType.UINT8,
        3: ModelDataType.INT8,
        4: ModelDataType.UINT16,
        5: ModelDataType.INT16,
        6: ModelDataType.INT32,
        7: ModelDataType.INT64,
        8: ModelDataType.STR,
        9: ModelDataType.BOOL,
        10: ModelDataType.FP16,
        11: ModelDataType.FP64,
        12: ModelDataType.UINT32,
        14: ModelDataType.UINT64,
        15: ModelDataType.COMPLEX128,
        16: ModelDataType.BFLOAT16
    }
    _str_map = {str(k):v.name for k, v in _map.items()}
    _inv_map = {v: k for k, v in _map.items()}
    _inv_name_map = {v.name: k for k, v in _map.items()}


@dataclass
class NP_map(Base_map):
    import numpy as np
    _map = {
        np.float32: ModelDataType.FP32,
        np.uint8: ModelDataType.UINT8,
        np.int8: ModelDataType.INT8,
        np.uint16: ModelDataType.UINT16,
        np.int16: ModelDataType.INT16,
        np.int32: ModelDataType.INT32,
        np.int64: ModelDataType.INT64,
        str: ModelDataType.STR,
        bool: ModelDataType.BOOL,
        np.float16: ModelDataType.FP16,
        np.float64: ModelDataType.FP64,
        np.uint32: ModelDataType.UINT32,
        np.uint64: ModelDataType.UINT64,
        np.complex128: ModelDataType.COMPLEX128
    }
    _str_map = {str(k):v.name for k, v in _map.items()}
    _inv_map = {v: k for k, v in _map.items()}
    _inv_name_map = {v.name: k for k, v in _map.items()}

if __name__ == '__main__':

    np_map = NP_map
    import numpy as np
    print(np_map.map(np.float32))



import os, glob
from dataclasses import dataclass
from typing import List

import numpy as np
import copy
from importlib import import_module
from onnc.forest.core.runtime import AbstractRuntimeImplementation

from onnc.forest.core.options import Options, RuntimeSettings
from onnc.bench.core.model import ModelFormat


@dataclass
class OpenvinoSettings(RuntimeSettings):
    pass


class OpenvinoOptions(Options):
    model_format = ModelFormat.OPENVINO_IRDIR

    @staticmethod
    def _ov_dir_to_xml(loadable):
        # If input type is OpenvinoIRDir, search for xml file inside it
        if os.path.isdir(loadable.src):
            xmls = list(
                filter(lambda x: x.endswith(".xml"),
                       glob.glob(f"{loadable.src}/*")))
            if len(xmls) != 1:
                raise RuntimeError(
                    f"OpenvinoIRDir should contain one xml file, found {len(xmls)}"
                )
            loadable.src = xmls[0]
        if not loadable.src.endswith(".xml"):
            raise RuntimeError(f"OpenvinoRuntime's input file should be .xml, "
                               f"got {loadable.src}")
        return loadable

    def __init__(self, loadable):
        self.settings: OpenvinoSettings = OpenvinoSettings()
        self.devices: List = ["CPU"]
        self.loadable = self._ov_dir_to_xml(loadable)


class OpenvinoRuntime(AbstractRuntimeImplementation):

    @staticmethod
    def is_me(options: Options) -> bool:
        return isinstance(options, OpenvinoOptions)

    def __init__(self):
        self.out_buf = None
        self._infer_request = None
        self._compiled_model = None
        self.__ov_runtime = import_module('openvino.runtime')

    def launch(self, options):
        self.core = self.__ov_runtime.Core()
        self.device = options.devices[0]

    def load(self, options):
        loadable = options.loadable
        self._output_names = [
            output.dump()["name"] for output in loadable.outputs
        ]
        assert loadable.src.endswith(".xml")
        model = self.core.read_model(loadable.src)

        # This step is belong to `materialize`
        # However, bind_input() need to use compiled_model
        # to create InferRequest, so, it is a workaround
        # and materialize the model in the step `load`
        self._compiled_model = self.core.compile_model(model, self.device)
        print("self._compiled_model",
              self._compiled_model.get_property("INFERENCE_NUM_THREADS"))

    def create_infer_request(self):

        infer_request = self._compiled_model.create_infer_request()
        return infer_request

    def _create_infer_request(self):
        from openvino.runtime.ie_api import InferRequest
        if isinstance(self._infer_request, InferRequest):
            return
        self._infer_request = self._compiled_model.create_infer_request()

    def bind_input(self, input, object):
        self._create_infer_request()

    def bind_output(self, output, object):
        self._create_infer_request()

    def bind_all_inputs(self):
        self._create_infer_request()

    def bind_all_outputs(self):
        self._create_infer_request()

    def materialize(self):
        pass

    def write(self, data):
        if not hasattr(self, '_compiled_model'):
            raise RuntimeError("Please run materialize before running infer")

        if not self._infer_request:
            raise RuntimeError("Please bind inputs and outputs")

        # self._infer_request = self._compiled_model.create_infer_request()
        # for all input node, copy input data
        for i, d in enumerate(data):
            input_tensor = self._infer_request.get_input_tensor(i)
            np.copyto(input_tensor.data, d)

    def run(self):
        if not self._infer_request:
            raise RuntimeError("Please call write() before run")
        self._infer_request.infer()

    def read(self):
        if self._output_names:
            return [
                self._infer_request.get_tensor(
                    self._compiled_model.output(name)).data
                for name in self._output_names
            ]
        # self._infer_request.get_tensor( self._compiled_model.output("bboxes")).data
        return copy.deepcopy(
            [output.data for output in self._infer_request.outputs])


# class OpenvinoAsyncRuntime(AbstractAsyncRuntimeImplementation):

#     def __init__(self):
#         pass

#     def launch(self, settings, devices):
#         pass

#     def load(self, model_path):
#         pass

#     def bind_input(self):
#         pass

#     def bind_output(self):
#         pass

#     def materialize(self):
#         pass

#     def submit(self, data):
#         pass

#     def wait(self):
#         """
#         Nothing to do here since start_async already implements this for us
#         """
#         pass

#     def outcome(self):
#         pass



from enum import Enum


class ModelFormat(Enum):
    NON_SPECIFIED = 0
    UNKNOWN = 1

    PT_NN_MODULE = 11
    TF_KERAS_MODEL = 12
    KERAS_MODEL = 13
    TF_SESSION = 14

    H5 = 21
    SAVED_MODEL = 22
    PB = 23
    ZIPPED_SAVED_MODEL = 24
    TFLITE = 25

    PTH = 31
    ONNX = 32
    TORCH_TRACED = 33

    CKPT = 26

    TRT_PLAN = 41
    OPENVINO_IRDIR = 42

    GRAPHDEF = 51

    RELAYIR = 61

    MLIR = 71


TF_MODEL_FORMATS = [ModelFormat.TF_KERAS_MODEL,
                    ModelFormat.KERAS_MODEL,
                    ModelFormat.H5,
                    ModelFormat.SAVED_MODEL,
                    ModelFormat.PB,
                    ModelFormat.TFLITE,
                    ModelFormat.CKPT]

PYTORCH_MODEL_FORMATS = [ModelFormat.PT_NN_MODULE,
                         ModelFormat.PTH,
                         ModelFormat.TORCH_TRACED]
class ModelDataType(Enum):
    NON_SPECIFIED = 0
    FP32 = 1
    UINT8 = 2
    INT8 = 3
    UINT16 = 4
    INT16 = 5
    INT32 = 6
    INT64 = 7
    STR = 8
    BOOL = 9
    FP16 = 10
    FP64 = 11
    UINT32 = 12
    UINT64 = 14
    COMPLEX128 = 15
    BFLOAT16 = 16



from pathlib import Path
from typing import Dict, List, Any, Callable, Union, Set
from pathlib import Path
import os
import time
import re
import zipfile
import sys
import json

import requests
from loguru import logger

from onnc.bench.core.deployment import Deployment
from .builder import IBuilder
from .saas_config import URI_MAP, timeout, poll_interval

from onnc.bench.core.common import get_tmp_path
from . import Compilation


def spinning_cursor():
    while True:
        for cursor in '|/-\\':
            yield cursor


class SaaSResult:

    def __init__(self, http_res):
        self.success = None
        self.message = None
        self.http_status_code = None
        self.data: Dict = {}

        self.from_http_res(http_res)

    def __bool__(self):
        return True if self.success else False

    def from_http_res(self, http_res) -> None:
        self.http_status_code = http_res.status_code

        if http_res.status_code == requests.codes.ok:
            self.success = True
            self.message = ""
            try:
                self.data = http_res.json()
            except Exception:
                self.data = {}
        else:
            self.success = False
            self.message = http_res.json()["error"]["message"]
            self.data = http_res.json()


class BearerAuth(requests.auth.AuthBase):

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class ONNCSaaSBuilder(IBuilder):
    BUILDER_NAME = "ONNCSaaSBuilder"

    def __init__(self, protocal: str, url: str, port: int):
        self.base_url = f"{protocal}://{url}:{port}"
        self.auth_token = ""
        self._compilations: Dict[int, Compilation] = {}
        self._project_id: str = ""
        self._build_id: str = ""

    def _http_req(self, api: Union[Callable, List], *args, **kwargs):
        if type(api) == list:
            uri = URI_MAP[api[0]]["uri"].format(*api[1:])
            method = URI_MAP[api[0]]["method"]
        elif callable(api):
            uri = URI_MAP[api.__name__]["uri"]
            method = URI_MAP[api.__name__]["method"]

        if method == 'GET':
            return requests.get(f"{self.base_url}{uri}",
                                auth=BearerAuth(self.auth_token),
                                *args,
                                **kwargs)
        elif method == 'POST':
            return requests.post(f"{self.base_url}{uri}",
                                 auth=BearerAuth(self.auth_token),
                                 *args,
                                 **kwargs)
        elif method == 'DELETE':
            return requests.delete(f"{self.base_url}{uri}",
                                   auth=BearerAuth(self.auth_token),
                                   *args,
                                   **kwargs)
        elif method == 'PATCH':
            return requests.patch(f"{self.base_url}{uri}",
                                  auth=BearerAuth(self.auth_token),
                                  *args,
                                  **kwargs)
        elif method == 'PUT':
            return requests.put(f"{self.base_url}{uri}",
                                auth=BearerAuth(self.auth_token),
                                *args,
                                **kwargs)

    def saas_login(self, email: str, password: str) -> SaaSResult:
        data = {"email": email, "password": password}
        r = self._http_req(self.saas_login, json=data)
        if r.status_code == requests.codes.ok:
            self.auth_token = r.json()["token"]
            logger.info(f"Login successfully")
        else:
            logger.error(f"Login failed")

        return SaaSResult(r)

    def saas_verify_key(self, api_key) -> bool:
        pass

    def saas_create_project(self, name: str, info: Dict = {}) -> SaaSResult:
        data = {"name": name, "info": info}
        r = self._http_req(self.saas_create_project, json=data)
        if r.status_code == requests.codes.ok:
            self._project_id = r.json()["id"]
            logger.debug(f"Project ID: {self._project_id}")

        return SaaSResult(r)

    def saas_get_project_id_by_name(self, project_name: str) -> SaaSResult:
        r = self._http_req(['saas_get_project_id_by_name', project_name])
        r_list = r.json()
        if len(r_list) > 0:
            self._project_id = r_list[0]["id"]
            logger.debug(f"Project ID: {self._project_id}")

        return SaaSResult(r)

    def saas_upload_file(self, file_path: str) -> SaaSResult:
        files = {os.path.basename(file_path): open(file_path, 'rb')}

        r = self._http_req(self.saas_upload_file, files=files)

        return SaaSResult(r)

    def _saas_create_compilation(self,
                                 model_path,
                                 dataset_path,
                                 params: Dict = None):
        if not params:
            params = {}
        logger.info(f'Start to upload model')
        logger.debug(f'Model path: {model_path}')
        model_sr = self.saas_upload_file(model_path)  # type: ignore[arg-type]
        logger.info('The model was uploaded successfully')

        logger.info(f'Start to upload dataset')
        logger.debug(f'Dataset path: {dataset_path}')
        dataset_sr = self.saas_upload_file(
            dataset_path)  # type: ignore[arg-type]
        logger.info('The dataset was uploaded successfully')

        data = {
            "model": 'oasis-fs://' + model_sr.data["files"][0]['id'],
            "modelSize": model_sr.data["files"][0]["size"],
            "calibration": 'oasis-fs://' + dataset_sr.data["files"][0]["id"],
            "calibrationSize": dataset_sr.data["files"][0]["size"],
            "compilerParameters": params
        }
        return data

    def saas_create_build(self, project_id: str, saas_compilations: List,
                          device: str) -> SaaSResult:
        data = {
            # "userId": "0",
            "projectId": project_id,
            "boardId": device,
            "input": saas_compilations
        }

        r = self._http_req(self.saas_create_build, json=data)

        sr = SaaSResult(r)
        logger.debug(f"Build ID: {sr.data['id']}")

        return sr

    def saas_get_build_state(self, build_id: str) -> SaaSResult:
        """Get state code of a build.

        :returns:
            The state code of the build
        :rtype:
            str
        """
        r = self._http_req(['saas_get_build_state', build_id])

        return SaaSResult(r)

    def saas_download_deployment_package(self, file_id, dest: Path):
        r = self._http_req(['saas_download_deployment_package', file_id],
                           stream=True)

        r.raise_for_status()

        logger.debug(f"Download destination: {dest}")

        with open(dest, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

        return SaaSResult(r)

    def saas_upload_report(self, compilation_id: str, report_type: str,
                           report: Dict) -> SaaSResult:
        data = {
            "values": report,
            "compilationId": compilation_id,
            "report_type": report_type
        }

        r = self._http_req(self.saas_upload_report, json=data)

        return SaaSResult(r)

    def saas_list_devices(self) -> SaaSResult:
        r = self._http_req(self.saas_list_devices)
        return SaaSResult(r)

    def build(self, target: str, converter_params=None) -> Dict:
        if not converter_params:
            params = {}
        saas_compilations = []

        # Upload files and create compilation
        for iternal_cid in self._compilations:
            compilation = self._compilations[iternal_cid]

            compilation.compiler_params['target'] = target

            params = {}
            params["model_meta"] = compilation.model_meta
            params["sample_meta"] = compilation.sample_meta
            params["compiler_params"] = compilation.compiler_params
            params["calibrator_params"] = compilation.calibrator_params
            params["converter_params"] = converter_params

            data = self._saas_create_compilation(compilation.model_path,
                                                 compilation.sample_path,
                                                 params)
            # saas_compilations.append(cr.data["id"])
            saas_compilations.append(data)

        # Create Build and trigger build/compilation
        build_sr = self.saas_create_build(project_id=self._project_id,
                                          saas_compilations=saas_compilations,
                                          device=target)
        if not build_sr:
            return build_sr.data

        self._build_id = build_sr.data["id"]

        logger.debug(f"ONNC-SAAS BuildID: {self._build_id}")

        # Wait and poll the result
        t = 0
        spinner = spinning_cursor()
        sys.stdout.write("Building... ")
        while t < (timeout / poll_interval):
            t += 1

            # update spinner
            sys.stdout.write(next(spinner))
            sys.stdout.flush()
            sys.stdout.write('\b')
            time.sleep(poll_interval)

            state_sr = self.saas_get_build_state(build_sr.data["id"])

            # HTTP failure
            if not state_sr:
                logger.error(f"Something was wrong: {state_sr.data}")
                return state_sr.data

            # compilation is still running
            elif state_sr and (state_sr.data["state"] in ["pending", "running"
                                                         ]):
                time.sleep(poll_interval)

            # compilation finishes
            else:
                if state_sr.data["state"] == "success":
                    logger.success(f"Compiled successfully.")
                    file_url = state_sr.data["output"]["deploymentPackage"]
                    logger.debug(file_url)
                else:
                    logger.error(f"Compiled Unsuccessfully: {state_sr.data}")

                return state_sr.data

        logger.error(f"Compilation failed: Timeout")
        return {"error": {"message": "Timeout"}, "id": self._build_id}

    def save(self, output: Path) -> Deployment:
        bs = self.saas_get_build_state(self._build_id)

        file_url = bs.data["output"]["deploymentPackage"]

        file_ids = re.findall('oasis-fs://(.*?)$', file_url)
        if len(file_ids) != 1:
            logger.error(f"Parse file_id error: {file_url}")
        else:
            file_id = file_ids[0]

        tmp_download = get_tmp_path() + '.zip'

        self.saas_download_deployment_package(file_id, tmp_download)

        logger.debug(f"Download file to {tmp_download}")

        with zipfile.ZipFile(tmp_download, 'r') as zip_ref:
            zip_ref.extractall(output)

        if bs.data["state"] == "success":
            deployment = Deployment(output)

        else:
            try:
                deployment = Deployment(None,
                                        bs.data["output"]["report"]["metrics"],
                                        bs.data["output"]["report"]["logs"])
            except KeyError:
                logger.error(f"Unable to parse SaaS result")
                logger.error(json.dumps(bs.data))
                deployment = Deployment(None, {}, [])

                return deployment

        return deployment

    @property
    def supported_devices(self) -> Dict:
        sr = self.saas_list_devices()
        devices = {x['name']: x['id'] for x in sr.data}
        return devices



from dataclasses import dataclass, field
import abc, typing
from typing import List, Union
from pathlib import Path

from onnc.bench.core.model.identifier import identify
from onnc.bench.core.model import ModelFormat
from onnc.bench.core.model.model import Model


@dataclass
class RuntimeSettings:
    pass


class OptionsRegistry(type):

    REGISTRY = []

    def __new__(cls, name, bases, attrs):
        new_cls = type.__new__(cls, name, bases, attrs)
        cls.REGISTRY.append(new_cls)
        return new_cls


class Options(metaclass=OptionsRegistry):
    model_format = None
    def __init__(self, loadable: Union[str, Path, Model]):
        self.settings: RuntimeSettings
        self.devices: List = []

        if isinstance(loadable, str) or isinstance(loadable, Path):
            loadable = Model(loadable)

        self.loadable = loadable
        self.__change_cls(loadable)

    def encapsulate(self, options):
        self._encapsulation.append(options)

    def __change_cls(self, loadable):
        format_ = identify(loadable)

        for opt_cls in OptionsRegistry.REGISTRY:
            if opt_cls.is_me(format_):
                """
                 The entire Options class is substituded by the derived class
                 after calling self.__class__ = opt_cls. Thus, the followeing
                 self.__init__ will call derived class's init
                """
                self.__class__ = opt_cls
                self.__init__(loadable)

                return

        raise NotImplementedError(f"Unable to identify {loadable}")

    @classmethod
    def is_me(cls, format) -> bool:
        return format == cls.model_format



from typing import List, Dict, Union
from pathlib import Path
import shutil
import json
import os
from os.path import abspath

from onnc.bench.core.model.model import Model


class Deployment:
    """
    Deployment is a container class for built artifacts and reports
    """
    META_FNAME = Path('.deployment.json')

    def __init__(self, path: Union[None, Path], report=None, logs=None):
        self._compiled_files: List = []
        self._report = report
        self._compile_logs = logs
        if path:
            self.base_path = Path(path)
            self.report_path = self.base_path / Path('report.json')
        else:
            self.base_path = Path("")
            self.report_path = Path('report.json')

    def __str__(self):
        try:
            with open(self.report_path, 'r') as f:
                report = json.load(f)
            return json.dumps(report, sort_keys=True, indent=2)
        except:
            return "{}"

    def __repr__(self):
        return json.dumps(self.meta, sort_keys=True, indent=2)

    @property
    def report(self) -> Dict:
        if not self._report:
            if os.path.exists(self.report_path):
                with open(self.report_path, 'r') as f:
                    return dict(json.load(f)["metrics"])
        else:
            return self._report

        return {}

    @property
    def compiled_files(self):
        if os.path.exists(self.base_path):
            model_src = self.base_path / Path('build')
            if model_src.exists():
                files = []
                # list all files recursively
                for i in os.walk(model_src):
                    if len(i[2]) > 0:
                        for j in i[2]:
                            files.append(os.path.join(i[0], j))
                return files

        return []

    @property
    def model_list(self) -> List[Model]:

        if not os.path.exists(self.base_path):
            raise Exception(f'Deployment base_path is not a directory: '
                            f'{self.base_path}.')
        res = []
        for sub_model_root in os.listdir(self.base_path):
            root = self.base_path / sub_model_root
            if not os.path.isdir(root / "build"):
                raise Exception(f'Invalid deployment subdir {root}: '
                                f'Missing model directory')
            model_to_add = None
            if os.path.isdir(root / "build/model"):
                """
                Case 1: submodel is a directory named model contains
                        multiple files. Ex: Openvino .xml, .bin
                """
                model_to_add = Model(abspath(root / "build/model"))
            else:
                """
                Case 2: submodel is a single file name starts with model
                """
                for file in os.listdir(root / "build"):
                    if file.startswith("model"):
                        model_to_add = Model(abspath(root / f"build/{file}"))
                        break
            if not model_to_add:
                raise Exception(
                    f"Unrecognized deployment subdir "
                    f"abspath{root}, plz contacts developers to fix it.")
            else:
                res.append(model_to_add)
        return res

    @property
    def compile_logs(self):
        if not self._compile_logs:
            if os.path.exists(self.report_path):
                with open(self.report_path, 'r') as f:
                    return json.load(f)["logs"]
        else:
            return self._compile_logs

        return []

    @property
    def meta(self):
        return {
            "base_path": str(self.base_path),
            "compiled_files": [str(x) for x in self.compiled_files],
            "report_path": str(self.report_path),
            "report": self.report
        }

    def save(self):
        _path = self.base_path / self.META_FNAME
        open(_path, 'w').write(json.dumps(self.meta, sort_keys=True, indent=4))

    def load(self):
        meta = json.loads(open(self.base_path / self.META_FNAME, 'r').read())

        self.base_path = meta["base_path"]
        self.report_path = meta["report_path"]

    def load_raw(self):
        """Scan folder and construct the object"""
        pass

    def deploy(self, target: Path):
        """Copy the deployment folder to target

        Copy the deployment folder to target and reconstruct the meta

        """
        shutil.copytree(self.base_path, target)
        if os.path.exists(target / self.META_FNAME):
            os.remove(target / self.META_FNAME)
        deployment = Deployment(target)

        return deployment
